from typing import Sequence, Optional, Literal, Union
from itertools import permutations
import math
import random
from votekit.ballot import Ballot, RankBallot
from votekit.pref_profile import RankProfile, ScoreProfile
import pandas as pd
import numpy as np
from numpy.typing import NDArray

COLOR_LIST = [
    "#0099cd",
    "#ffca5d",
    "#00cd99",
    "#99cd00",
    "#cd0099",
    "#9900cd",
    "#8dd3c7",
    "#bebada",
    "#fb8072",
    "#80b1d3",
    "#fdb462",
    "#b3de69",
    "#fccde5",
    "#bc80bd",
    "#ccebc5",
    "#ffed6f",
    "#ffffb3",
    "#a6cee3",
    "#1f78b4",
    "#b2df8a",
    "#33a02c",
    "#fb9a99",
    "#e31a1c",
    "#fdbf6f",
    "#ff7f00",
    "#cab2d6",
    "#6a3d9a",
    "#b15928",
    "#64ffda",
    "#00B8D4",
    "#A1887F",
    "#76FF03",
    "#DCE775",
    "#B388FF",
    "#FF80AB",
    "#D81B60",
    "#26A69A",
    "#FFEA00",
    "#6200EA",
]


def ballots_by_first_cand(profile: RankProfile) -> dict[str, list[RankBallot]]:
    """
    Partitions the profile by first place candidate. Assumes there are no ties within first place
    positions of ballots.

    Args:
        profile (RankProfile): Profile to partititon.

    Returns:
        dict[str, list[RankBallot]]:
            A dictionary whose keys are candidates and values are lists of ballots that
            have that candidate first.
    """
    if not isinstance(profile, RankProfile):
        raise TypeError("Ballots must have rankings.")

    df = profile.df
    assert profile.max_ranking_length is not None
    ranking_cols = [f"Ranking_{i}" for i in range(1, profile.max_ranking_length + 1)]

    rank_arr = df[ranking_cols].to_numpy()
    weights = df["Weight"].to_numpy()
    voter_sets = df["Voter Set"].to_numpy().astype(object)

    cand_dict: dict[str, list[RankBallot]] = {c: [] for c in profile.candidates}
    tilde = frozenset({"~"})

    for row, w, voter_set in zip(rank_arr, weights, voter_sets):
        first = row[0]

        if len(first) > 1:
            raise ValueError(
                f"Ballot "
                f"{RankBallot(ranking=tuple(c_set for c_set in row if c_set != tilde), weight=float(w))} "
                "has a tie for first."
            )

        cand = next(iter(first))

        if cand == "~":
            continue

        clean_ranking = tuple(s for s in row if s != tilde)

        cand_dict[cand].append(
            RankBallot(ranking=clean_ranking, weight=float(w), voter_set=voter_set)
        )

    return cand_dict


def add_missing_cands(profile: RankProfile) -> RankProfile:
    """
    Add any candidates from `profile.candidates` that are not listed on a ballot
    as tied in last place.

    Args:
        profile (RankProfile): Input profile.

    Returns:
        RankProfile
    """
    if not isinstance(profile, RankProfile):
        raise TypeError("Profile must be of type RankProfile.")
    new_ballots = [RankBallot()] * len(profile.ballots)
    candidates = set(profile.candidates)

    for i, ballot in enumerate(profile.ballots):
        if ballot.ranking is None:
            raise TypeError("Ballots must have rankings.")
        else:
            b_cands = [c for s in ballot.ranking for c in s]
            missing_cands = candidates.difference(b_cands)

            new_ranking = (
                list(ballot.ranking) + [missing_cands]
                if len(missing_cands) > 0
                else ballot.ranking
            )

            new_ballots[i] = RankBallot(
                weight=ballot.weight,
                voter_set=ballot.voter_set,
                ranking=tuple([frozenset(s) for s in new_ranking]),
            )

    return RankProfile(ballots=tuple(new_ballots), candidates=tuple(candidates))


def validate_score_vector(score_vector: Sequence[float]):
    """
    Validator function for score vectors. Vectors should be non-increasing and non-negative.

    Args:
        score_vector (Sequence[float]): Score vector.

    Raises:
        ValueError: If any score is negative.
        ValueError: If score vector is increasing at any point.

    """

    for i, score in enumerate(score_vector):
        # if score is negative
        if score < 0:
            raise ValueError("Score vector must be non-negative.")

        if i > 0:
            # if the current score is bigger than prev
            if score > score_vector[i - 1]:
                raise ValueError("Score vector must be non-increasing.")


def _score_dict_from_rankings_df_no_ties(
    profile: RankProfile,
    score_vector: Sequence[float],
) -> dict[str, float]:

    validate_score_vector(score_vector)

    if not isinstance(profile, RankProfile):
        raise TypeError("Profile must only contain ranked ballots.")

    assert profile.max_ranking_length is not None
    max_len = profile.max_ranking_length
    if len(score_vector) < max_len:
        score_vector = list(score_vector) + [0] * (max_len - len(score_vector))

    df = profile.df

    cand_frznst = [frozenset({c}) for c in profile.candidates_cast]
    all_frznst = cand_frznst + [frozenset({"~"}), frozenset()]
    n_buckets = len(all_frznst)
    idx_of_empty = all_frznst.index(frozenset())

    # Pull out weights and score_vector as NumPy arrays:
    weights = df["Weight"].to_numpy(dtype=float)
    score_arr = np.array(score_vector, dtype=float).reshape(1, -1)

    rank_cols = [f"Ranking_{i}" for i in range(1, max_len + 1)]
    arr = df[rank_cols].to_numpy(dtype=object)
    flat_arr = arr.ravel()

    # Slick way of converting frozensets to integer codes:
    codes_flat: NDArray[np.int64] = pd.Categorical(
        flat_arr, categories=all_frznst
    ).codes.astype(np.int64)

    # Take care of error codes (-1)
    if (codes_flat == -1).any():
        codes_flat = np.where(codes_flat == -1, idx_of_empty, codes_flat)

    weight_matrix = weights[:, None] * score_arr

    weights_flat = weight_matrix.ravel()
    bucket_sums = np.bincount(codes_flat, weights=weights_flat, minlength=n_buckets)

    return {next(iter(k)): bucket_sums[idx] for idx, k in enumerate(cand_frznst)}


def score_dict_from_score_vector(
    profile: RankProfile,
    score_vector: Sequence[float],
    tie_convention: Literal["high", "average", "low"] = "low",
) -> dict[str, float]:
    """
    Score the candidates based on a score vector. For example, the vector (1,0,...) would
    return the first place votes for each candidate. Vectors should be non-increasing and
    non-negative. Vector should be as long as ``max_ranking_length`` in the profile.
    If it is shorter, we add 0s. Candidates who are not mentioned in any ranking do not appear
    in the dictionary.


    Args:
        profile (RankProfile): Profile to score.
        score_vector (Sequence[float]): Score vector. Should be
            non-increasing and non-negative. Vector should be as long as ``max_ranking_length`` in
            the profile. If it is shorter, we add 0s.
        tie_convention (Literal["high", "average", "low"], optional): How to award points for
            tied rankings. Defaults to "low", where any candidates tied receive the lowest possible
            points for their position, eg three people tied for 3rd would each receive the points
            for 5th. "high" awards the highest possible points, so in the previous example, they
            would each receive the points for 3rd. "average" averages the points, so they would each
            receive the points for 4th place.

    Returns:
        dict[str, float]:
            Dictionary mapping candidates to scores.
    """
    validate_score_vector(score_vector)

    if not isinstance(profile, RankProfile):
        raise TypeError("Profile must only contain ranked ballots.")
    assert profile.max_ranking_length is not None
    max_length = profile.max_ranking_length
    if len(score_vector) < max_length:
        score_vector = list(score_vector) + [0] * (max_length - len(score_vector))

    scores = {c: 0.0 for c in profile.candidates_cast}

    try:
        ranking_cols = [f"Ranking_{i}" for i in range(1, max_length + 1)]
        ranking_mat = profile.df[ranking_cols].to_numpy()
    except KeyError as e:
        raise TypeError("Ballots must have rankings.") from e

    weights = profile.df["Weight"].to_numpy(dtype=float)

    if tie_convention not in ["high", "average", "low"]:
        raise ValueError(
            (
                "tie_convention must be one of 'high', 'low', 'average', "
                f"not {tie_convention}"
            )
        )

    tilde = frozenset({"~"})
    for idx in range(len(ranking_mat)):
        current_ind = 0
        ranking = ranking_mat[idx]
        wt = weights[idx]
        for s in ranking:
            position_size = len(s)
            if position_size == 0:
                raise TypeError(
                    f"Ballot {Ballot(ranking=ranking.tolist(), weight=wt)} has an empty ranking "
                    "position."
                )
            if s == tilde:
                continue

            local_score_vector: Sequence[float | int] = score_vector[
                current_ind : current_ind + position_size
            ]

            if tie_convention == "high":
                allocation = max(local_score_vector)
            elif tie_convention == "low":
                allocation = min(local_score_vector)
            else:
                allocation = sum(local_score_vector) / position_size

            for c in s:
                scores[c] += allocation * wt
            current_ind += position_size

    return scores


def _first_place_votes_from_df_no_ties(
    profile: RankProfile,
) -> dict[str, float]:
    """
    Computes first place votes for all candidates_cast in a ``RankProfile``.
    Intended to be much faster than first_place_votes, but does not handle ties in ballots.

    Args:
        profile (RankProfile): The profile to compute first place votes for.

    Returns:
        dict[str, float]:
            Dictionary mapping candidates to number of first place votes.
    """
    # equiv to score vector of (1,0,0,...)
    assert profile.max_ranking_length is not None
    return _score_dict_from_rankings_df_no_ties(
        profile, [1] + [0] * (profile.max_ranking_length - 1)
    )


def first_place_votes(
    profile: RankProfile,
    tie_convention: Literal["high", "average", "low"] = "average",
) -> dict[str, float]:
    """
    Computes first place votes for all candidates_cast in a ``RankProfile``.

    Args:
        profile (RankProfile): The profile to compute first place votes for.
        tie_convention (Literal["high", "average", "low"], optional): How to award points
            for tied first place votes. Defaults to "average", where if n candidates are tied for
            first, each receives 1/n points. "high" would award them each one point, and "low" 0.

    Returns:
        dict[str, float]:
            Dictionary mapping candidates to number of first place votes.
    """
    # equiv to score vector of (1,0,0,...)
    if not isinstance(profile, RankProfile):
        raise TypeError("Profile must be of type RankProfile.")
    assert profile.max_ranking_length is not None
    return score_dict_from_score_vector(
        profile, [1] + [0] * (profile.max_ranking_length - 1), tie_convention
    )


def mentions(
    profile: RankProfile,
) -> dict[str, float]:
    """
    Calculates total mentions for all candidates in a ``RankProfile``.

    Args:
        profile (RankProfile): RankProfile of ballots.

    Returns:
        dict[str, float]:
            Dictionary mapping candidates to mention totals (values).
    """
    mentions = {c: 0.0 for c in profile.candidates}
    if not isinstance(profile, RankProfile):
        raise TypeError("Profile must be of type RankProfile.")
    for ballot in profile.ballots:
        if ballot.ranking is None:
            raise TypeError("Ballots must have rankings.")
        else:
            for s in ballot.ranking:
                for cand in s:
                    mentions[cand] += ballot.weight
    return mentions


def borda_scores(
    profile: RankProfile,
    borda_max: Optional[int] = None,
    tie_convention: Literal["high", "average", "low"] = "low",
) -> dict[str, float]:
    r"""
    Calculates Borda scores for candidates_cast in a ``RankProfile``. The Borda vector is
    :math:`(n,n-1,\dots,1)` where :math:`n` is the ``borda_max`.

    Args:
        profile (RankProfile): ``RankProfile`` of ballots.
        borda_max (int, optional): The maximum value of the Borda vector. Defaults to
            the length of the longest allowable ballot in the profile.
        tie_convention (Literal["high", "average", "low"], optional): How to award points for
            tied rankings. Defaults to "low", where any candidates tied receive the lowest possible
            points for their position, eg three people tied for 3rd would each receive the points
            for 5th. "high" awards the highest possible points, so in the previous example, they
            would each receive the points for 3rd. "average" averages the points, so they would each
            receive the points for 4th place.

    Returns:
        dict[str, float]:
            Dictionary mapping candidates to Borda scores.
    """
    if not isinstance(profile, RankProfile):
        raise TypeError("Profile must be of type RankProfile.")
    if borda_max is None:
        assert profile.max_ranking_length is not None
        borda_max = profile.max_ranking_length

    score_vector = list(range(borda_max, 0, -1))

    return score_dict_from_score_vector(profile, score_vector, tie_convention)


def tiebreak_set(
    r_set: frozenset[str],
    profile: Optional[RankProfile] = None,
    tiebreak: str = "random",
    scoring_tie_convention: Literal["high", "average", "low"] = "low",
    backup_tiebreak_convention: Optional[str] = None,
) -> tuple[frozenset[str], ...]:
    """
    Break a single set of candidates into multiple sets each with a single candidate according
    to a tiebreak rule. Rule 1: random. Rule 2: first-place votes; break the tie based on
    first-place votes in the profile. Rule 3: borda; break the tie based on Borda points in the
    profile. Rule 4: lex/lexicographic/alph/alphabetical; break the tie alphabetically.

    Args:
        r_set (frozenset[str]): Set of candidates on which to break tie.
        profile (RankProfile, optional): Profile used to break ties in first-place votes or
            Borda setting. Defaults to None, which implies a random tiebreak.
        tiebreak (str): Tiebreak method to use. Options are "random", "first_place", and
            "borda". Defaults to "random".
        scoring_tie_convention (Literal["high", "average", "low"]): How to award points
            for tied rankings. Defaults to "low", where any candidates tied receive the lowest
            possible points for their position, eg three people tied for 3rd would each receive the
            points for 5th. "high" awards the highest possible points, so in the previous example,
            they would each receive the points for 3rd. "average" averages the points, so they would
            each receive the points for 4th place.
        backup_tiebreak_convention (str, optional): If the initial tiebreak does not resolve all
            ties,
            this convention is used to break any remaining ties. Options are "random" and
            "lex/lexicographic/alph/alphabetical". Defaults to None which sets the backup to
            "lex" if the initial tiebreak is alphabetical, and "random" otherwise.

    Returns:
        tuple[frozenset[str],...]: tiebroken ranking
    """
    if tiebreak in ["alphabetical", "lexicographic", "alph", "lex"]:
        sorted_cands = sorted([c for c in r_set])
        new_ranking = tuple(map(lambda c: frozenset({c}), sorted_cands))

    elif tiebreak == "random":
        new_ranking = tuple(
            frozenset({c}) for c in random.sample(list(r_set), k=len(r_set))
        )
    elif (tiebreak == "first_place" or tiebreak == "borda") and profile:
        if tiebreak == "borda":
            tiebreak_scores = borda_scores(
                profile, tie_convention=scoring_tie_convention
            )
        else:
            tiebreak_scores = first_place_votes(
                profile, tie_convention=scoring_tie_convention
            )
        tiebreak_scores = {
            c: score for c, score in tiebreak_scores.items() if c in r_set
        }
        new_ranking = score_dict_to_ranking(tiebreak_scores)

    elif profile is None:
        raise ValueError("Method of tiebreak requires profile.")
    else:
        raise ValueError("Invalid tiebreak code was provided")

    if any(len(s) > 1 for s in new_ranking):
        if backup_tiebreak_convention is None:
            if tiebreak in [
                "alphabetical",
                "lexicographic",
                "alph",
                "lex",
            ]:
                backup_tiebreak_convention = "lex"
            else:
                backup_tiebreak_convention = "random"

        if backup_tiebreak_convention in [
            "alphabetical",
            "lexicographic",
            "alph",
            "lex",
        ]:
            print("Initial tiebreak was unsuccessful, performing alphabetic tiebreak")
            new_ranking, _ = tiebroken_ranking(
                new_ranking, profile=profile, tiebreak="lex"
            )
        elif backup_tiebreak_convention == "random":
            print("Initial tiebreak was unsuccessful, performing random tiebreak")
            new_ranking, _ = tiebroken_ranking(
                new_ranking, profile=profile, tiebreak="random"
            )
        else:
            raise ValueError("Invalid backup tiebreak code was provided")

    return new_ranking


def tiebroken_ranking(
    ranking: tuple[frozenset[str], ...],
    profile: Optional[RankProfile] = None,
    tiebreak: str = "random",
) -> tuple[
    tuple[frozenset[str], ...], dict[frozenset[str], tuple[frozenset[str], ...]]
]:
    """
    Breaks ties in a list-of-sets ranking according to a given scheme.

    Args:
        ranking (list[set[str]]): A list-of-set ranking of candidates.
        profile (RankProfile, optional): Profile used to break ties in first-place votes or
            Borda setting. Defaults to None, which implies a random tiebreak.
        tiebreak (str, optional): Method of tiebreak, currently supports 'random', 'borda',
            'first_place'. Defaults to random.

    Returns:
        tuple[tuple[frozenset[str], ...], dict[frozenset[str], tuple[frozenset[str],...]]]:
            The first entry of the tuple is a list-of-set ranking of candidates (broken down to one
            candidate sets). The second entry is a dictionary that maps tied sets to their
            resolution.
    """
    new_ranking: list[frozenset[str]] = [frozenset()] * len(
        [c for s in ranking for c in s]
    )

    i = 0
    tied_dict = {}
    for s in ranking:
        if len(s) > 1:
            tiebroken = tiebreak_set(s, profile, tiebreak)
            tied_dict[s] = tiebroken
        else:
            tiebroken = (s,)
        new_ranking[i : (i + len(tiebroken))] = tiebroken
        i += len(tiebroken)

    return (tuple(new_ranking), tied_dict)


def score_dict_to_ranking(
    score_dict: dict[str, float], sort_high_low: bool = True
) -> tuple[frozenset[str], ...]:
    """
    Sorts candidates into a tuple of frozensets ranking based on a scoring dictionary.

    Args:
        score_dict (dict[str, float]): Dictionary between candidates
            and their score.
        sort_high_low (bool, optional): How to sort candidates based on scores. True sorts
            from high to low. Defaults to True.


    Returns:
        tuple[frozenset[str],...]: Candidate rankings in a list-of-sets form.
    """

    score_to_cand: dict[float, list[str]] = {s: [] for s in score_dict.values()}
    for c, score in score_dict.items():
        score_to_cand[score].append(c)

    if len(score_to_cand) > 0:
        return tuple(
            [
                frozenset(c_list)
                for _, c_list in sorted(
                    score_to_cand.items(), key=lambda x: x[0], reverse=sort_high_low
                )
            ]
        )
    else:
        return (frozenset(),)


def elect_cands_from_set_ranking(
    ranking: Sequence[Union[frozenset[str], set[str]]],
    m: int,
    profile: Optional[RankProfile] = None,
    tiebreak: Optional[str] = None,
) -> tuple[
    tuple[frozenset[str], ...],
    tuple[frozenset[str], ...],
    Optional[tuple[frozenset[str], tuple[frozenset[str], ...]]],
]:
    """
    Given a ranking, elect the top m candidates in the ranking.
    If a tie set overlaps the desired number of seats, it breaks the tie with the provided
    method or raises a ValueError if tiebreak is set to None.
    Returns a tuple of elected candidates, remaining candidates, and a tuple whose first entry
    is a tie set and whose second entry is the resolution of the tie.

    Args:
        ranking (tuple[frozenset[str],...]): A list-of-set ranking of candidates.
        m (int): Number of seats to elect.
        profile (RankProfile, optional): Profile used to break ties in first-place votes or
            Borda setting. Defaults to None, which implies a random tiebreak.
        tiebreak (str, optional): Method of tiebreak, currently supports 'random', 'borda',
            'first_place'. Defaults to None, which does not break ties.

    Returns:
        tuple[tuple[frozenset[str]]], list[tuple[frozenset[str]],
            Optional[tuple[frozenset[str], tuple[frozenset[str], ...]]]:
            A list-of-sets of elected candidates, a list-of-sets of remaining candidates,
            and a tuple whose first entry is a tie set and whose second entry is the resolution of
            the tie. If no ties were broken, the tuple returns None.
    """
    if m < 1:
        raise ValueError("m must be strictly positive")
    if m > len([c for s in ranking for c in s]):
        raise ValueError("m must be no more than the number of candidates.")

    ranking_fs: tuple[frozenset[str], ...] = tuple(
        s if isinstance(s, frozenset) else frozenset(s) for s in ranking
    )

    num_elected = 0
    elected: list[frozenset[str]] = []
    i = 0
    tiebreak_ranking: Optional[tuple[frozenset[str], tuple[frozenset[str], ...]]] = None

    while num_elected < m:
        elected.append(ranking_fs[i])
        num_elected += len(ranking_fs[i])
        if num_elected > m:
            if tiebreak is None:
                raise ValueError(
                    "Cannot elect correct number of candidates without breaking ties."
                )
            # back out the overfill
            elected.pop()
            num_elected -= len(ranking_fs[i])

            tiebroken = tiebreak_set(frozenset(ranking_fs[i]), profile, tiebreak)
            elected += tiebroken[: (m - num_elected)]

            remaining: list[frozenset[str]] = list(tiebroken[(m - num_elected) :])
            if i < len(ranking_fs):
                remaining += list(ranking_fs[(i + 1) :])

            return (
                tuple(elected),
                tuple(remaining),
                (ranking_fs[i], tiebroken),
            )
        i += 1

    return (tuple(elected), ranking_fs[i:], tiebreak_ranking)


def expand_tied_ballot(ballot: RankBallot) -> list[RankBallot]:
    """
    Fix tie(s) in a ballot by returning all possible permutations of the tie(s), and divide the
    weight of the original ballot equally among the new ballots.

    Args:
        ballot (RankBallot): Ballot to expand tie sets on.

    Returns:
        list[v]: All possible permutations of the tie(s).

    """
    if not isinstance(ballot, RankBallot):
        raise TypeError("Ballot must be of type RankBallot.")
    if ballot.ranking is None:
        raise TypeError("Ballot must have ranking.")
    if all(len(s) == 1 for s in ballot.ranking):
        return [ballot]

    else:
        for i, s in enumerate(ballot.ranking):
            if len(s) > 1:
                new_ballots = [
                    RankBallot(
                        weight=ballot.weight / math.factorial(len(s)),
                        voter_set=ballot.voter_set,
                        ranking=tuple(ballot.ranking[:i])
                        + tuple([frozenset({c}) for c in order])
                        + tuple(ballot.ranking[(i + 1) :]),
                    )
                    for order in permutations(s)
                ]

                return [b for new_b in new_ballots for b in expand_tied_ballot(new_b)]

        assert False  # mypy


def resolve_profile_ties(profile: RankProfile) -> RankProfile:
    """
    Takes in a PeferenceProfile with potential ties in ballots. Replaces
    ballots with ties with fractionally weighted ballots corresponding to
    all permutations of the tied ranking.

    Args:
        profile (RankProfile): Input profile with potentially tied rankings.

    Returns:
        RankProfile: A RankProfile with resolved ties.
    """

    new_ballots = tuple(
        [b for ballot in profile.ballots for b in expand_tied_ballot(ballot)]
    )
    return RankProfile(ballots=new_ballots)


def score_profile_from_ballot_scores(
    profile: ScoreProfile,
) -> dict[str, float]:
    """
    Score the candidates based on the ``scores`` parameter of the ballots.
    All ballots must have a ``scores`` parameter; note that a ``scores`` dictionary
    with no non-zero scores will raise the same error.

    Args:
        profile (ScoreProfile): Profile to score.

    Returns:
        dict[str, float]:
            Dictionary mapping candidates to scores.
    """
    scores = {c: 0.0 for c in profile.candidates}
    if not isinstance(profile, ScoreProfile):
        raise TypeError("Profile must be of type ScoreProfile.")
    for ballot in profile.ballots:
        if ballot.scores is None:
            raise TypeError(f"Ballot {ballot} has no scores.")
        else:
            for c, score in ballot.scores.items():
                scores[c] += score * ballot.weight

    return scores


def ballot_lengths(profile: RankProfile) -> dict[int, float]:
    """
    Compute the frequency of ballot lengths in the profile.
    Includes all lengths from 1 to ``max_ranking_length`` as keys.
    Ballots must have rankings.

    Args:
        profile (RankProfile): Profile to compute ballot lengths.

    Returns:
        dict[int, float]: Dictionary of ballot length frequency.

    Raises:
        TypeError: All ballots must have rankings.
    """
    if not isinstance(profile, RankProfile):
        raise TypeError("Profile must be of type RankProfile.")
    assert profile.max_ranking_length is not None

    ballot_lengths = {i: 0.0 for i in range(1, profile.max_ranking_length + 1)}

    for ballot in profile.ballots:
        if ballot.ranking is None:
            raise TypeError("All ballots must have rankings.")

        length = len(ballot.ranking)
        ballot_lengths[length] += ballot.weight

    return ballot_lengths


def index_to_lexicographic_ballot(
    ballot_index: int,
    n_candidates: int,
    max_length: int,
    total_valid_ballots_method,
    always_use_total_valid_ballots_method: bool = True,
) -> list[int]:
    """
    Convert an index to one ballot with candidates taken from the list range(n_candidates), and
    where the ballot has length at most max_length.
    The ordering of the ballots is lexicographic, i.e., the first ballot is the
    lexicographically smallest ballot and continues in that order:

    (0,),
    (0,1),
    (0,1,2),
    ...
    (0,2),
    (0,2,1),
    ...
    (n-1, n-2, ..., n-l)

    where n is the number of candidates and l is the maximum ballot length.

    Args:
        ballot_index (int): The index to convert.
        n_candidates (int): The number of candidates.
        max_length (int): The maximum allowed ballot rank.
        total_valid_ballots_method: ((n_candidates, max_length) ->
            integer) a method which returns the total number of
            allowed ballots.
        always_use_total_valid_ballots_method (bool): a flag
            indicating whether the given total valid ballots method
            should be used when computing b_n. If
            total_valid_ballots_method is using a cache then this
            should be True for maximum runtime efficiency.

    Returns:
        list[int]: A list representing the ballot corresponding to index.
    """
    total_valid_ballots = total_valid_ballots_method(n_candidates, max_length)
    if ballot_index >= total_valid_ballots:
        raise Exception(
            f"Given ballot index {ballot_index} out of range. Max index: {total_valid_ballots}"
        )

    def chunk_size(n_cands: int, ballot_len: int) -> int:
        return total_valid_ballots_method(n_cands, ballot_len) // n_cands

    candidates = list(range(n_candidates))
    out = []
    if always_use_total_valid_ballots_method:
        bn = total_valid_ballots_method(n_candidates + 1, max_length + 1) // (
            n_candidates + 1
        )
    else:
        bn = chunk_size(n_candidates + 1, max_length + 1)

    for i in range(n_candidates, 0, -1):
        if always_use_total_valid_ballots_method:
            bn = total_valid_ballots_method(
                n_candidates=i, max_ballot_length=max_length
            ) // (i)
        else:
            bn = (bn - 1) // i
        # Perform Euclidean division of index by bn
        section = ballot_index // bn
        remaining = ballot_index % bn
        out.append(candidates.pop(section))
        if remaining == 0:
            # Cut off the ballot here
            break
        ballot_index = remaining - 1
    return out


def build_df_from_ballot_samples(
    ballots_freq_dict: dict[tuple[int, ...], int], candidates: Sequence[str]
):
    """
    Helper function which creates a pandas df to instantiate a
    RankProfile
    args:
        ballots_freq_dict: dictionary mapping ballots to
            sampled frequency. The keys should be in candidate id
            form
        candidates : list of candidates in the profile
    returns:
        pandas df
    """
    df_data = []
    n_cands = len(candidates)
    for ballot in ballots_freq_dict.keys():
        ballot_as_frozenset_entries = tuple(
            [frozenset([candidates[i]]) for i in ballot]
        )
        completed_ballot = (
            ballot_as_frozenset_entries
            + tuple(
                [frozenset(["~"]) for _ in range(n_cands - len(ballot))]
            )  # padding short ballots
            + tuple([ballots_freq_dict[ballot], set()])
        )  # weight, voter set
        df_data.append(completed_ballot)
    return pd.DataFrame(
        df_data,
        columns=[f"Ranking_{i}" for i in range(1, n_cands + 1)]
        + ["Weight", "Voter Set"],
    )



"""
Generate ranked preference profiles using the Cambridge model.

The main API functions in this module are:

- `cambridge_profile_generator`: Generates a single preference profile using the Cambridge model.
- `cambridge_profiles_by_bloc_generator`: Generates preference profiles for each bloc using the
    Cambridge model.
"""

import numpy as np
from pathlib import Path
import pickle
import random
from typing import Optional
import apportionment.methods as apportion

from votekit.ballot import RankBallot
from votekit.pref_profile import RankProfile
from votekit.ballot_generator.bloc_slate_generator.model import BlocSlateConfig

# ===========================================================
# ================= Interior Work Functions =================
# ===========================================================


def _inner_cambridge_sampler(
    config: BlocSlateConfig,
    path: Path,
    majority_bloc: str,
    minority_bloc: str,
    historical_majority: str,
    historical_minority: str,
) -> dict[str, RankProfile]:
    """
    Inner function to generate profiles by bloc using Cambridge model.

    Args:
        config (BlocSlateConfig): Configuration object containing all necessary parameters for
            working with a bloc-slate ballot generator.
        path (Path): File path to an election data file to sample from.
        majority_bloc (str): Name of the bloc corresponding to the majority bloc.
        minority_bloc (str): Name of the bloc corresponding to the minority bloc.
        historical_majority (str): Name of the bloc in the historical data corresponding to the majority
            bloc in the current configuration.
        historical_minority (str): Name of the bloc in the historical data corresponding to the minority
            bloc in the current configuration.

    Returns:
        dict[str, RankProfile]: A dictionary whose keys are bloc strings and values are
            ``RankProfile`` objects representing the generated preference profiles for each bloc.
    """

    bloc_to_historical = {
        majority_bloc: historical_majority,
        minority_bloc: historical_minority,
    }

    with open(path, "rb") as pickle_file:
        ballot_frequencies = pickle.load(pickle_file)

    cohesion_parameters = {b: config.cohesion_df[b].loc[b] for b in config.blocs}

    # compute the number of bloc and crossover voters in each bloc using Huntington Hill
    voter_types = [
        (b, t) for b in list(config.bloc_proportions.keys()) for t in ["bloc", "cross"]
    ]

    voter_props = [
        (
            cohesion_parameters[b] * config.bloc_proportions[b]
            if t == "bloc"
            else (1 - cohesion_parameters[b]) * config.bloc_proportions[b]
        )
        for b, t in voter_types
    ]

    ballots_per_type = {
        k: int(v)
        for k, v in zip(
            voter_types,
            apportion.compute("huntington", voter_props, config.n_voters),  # type: ignore
        )
    }

    bloc_lst = config.blocs

    bloc_counts = apportion.compute(
        "huntington", list(config.bloc_proportions.values()), config.n_voters
    )
    if not isinstance(bloc_counts, list):
        if not isinstance(bloc_counts, int):
            raise TypeError(
                f"Unexpected type from apportionment got {type(bloc_counts)}"
            )

        bloc_counts = [bloc_counts]

    pp_by_bloc = {b: RankProfile() for b in bloc_lst}

    # FIX: Change this to use blocs and slates
    for i, bloc in enumerate(bloc_lst):
        bloc_voters = ballots_per_type[(bloc, "bloc")]
        cross_voters = ballots_per_type[(bloc, "cross")]
        ballot_pool = [RankBallot()] * (bloc_voters + cross_voters)

        opp_bloc = bloc_lst[(i + 1) % 2]

        bloc_first_count = sum(
            [
                freq
                for ballot, freq in ballot_frequencies.items()
                if ballot[0] == bloc_to_historical[bloc]
            ]
        )

        opp_bloc_first_count = sum(
            [
                freq
                for ballot, freq in ballot_frequencies.items()
                if ballot[0] == bloc_to_historical[opp_bloc]
            ]
        )

        pref_interval_dict = config.get_combined_preference_intervals_by_bloc()[bloc]

        # compute the relative probabilities of each ballot
        # sorted by ones where the ballot lists the bloc first
        # and those that list the opp first
        prob_ballot_given_bloc_first = {
            ballot: freq / bloc_first_count
            for ballot, freq in ballot_frequencies.items()
            if ballot[0] == bloc_to_historical[bloc]
        }

        prob_ballot_given_opp_first = {
            ballot: freq / opp_bloc_first_count
            for ballot, freq in ballot_frequencies.items()
            if ballot[0] == bloc_to_historical[opp_bloc]
        }

        bloc_voter_ordering = random.choices(
            list(prob_ballot_given_bloc_first.keys()),
            weights=list(prob_ballot_given_bloc_first.values()),
            k=bloc_voters,
        )
        cross_voter_ordering = random.choices(
            list(prob_ballot_given_opp_first.keys()),
            weights=list(prob_ballot_given_opp_first.values()),
            k=cross_voters,
        )

        for i in range(bloc_voters + cross_voters):
            # Based on first choice, randomly choose
            # ballots weighted by Cambridge frequency
            if i < bloc_voters:
                bloc_ordering = bloc_voter_ordering[i]
            else:
                bloc_ordering = cross_voter_ordering[i - bloc_voters]

            pl_ordering = list(
                np.random.choice(
                    list(pref_interval_dict.interval.keys()),
                    len(pref_interval_dict.interval),
                    p=list(pref_interval_dict.interval.values()),
                    replace=False,
                )
            )
            ordered_bloc_slate = [
                c for c in pl_ordering if c in config.slate_to_candidates[bloc]
            ]
            ordered_opp_slate = [
                c for c in pl_ordering if c in config.slate_to_candidates[opp_bloc]
            ]

            # Fill in the bloc slots as determined
            # With the candidate ordering generated with PL
            full_ballot = []
            for b in bloc_ordering:
                if b == bloc_to_historical[bloc]:
                    if ordered_bloc_slate:
                        full_ballot.append(ordered_bloc_slate.pop(0))
                else:
                    if ordered_opp_slate:
                        full_ballot.append(ordered_opp_slate.pop(0))

            ranking = tuple([frozenset({cand}) for cand in full_ballot])
            ballot_pool[i] = RankBallot(ranking=ranking, weight=1)

        pp = RankProfile(ballots=tuple(ballot_pool))
        pp = pp.group_ballots()
        pp_by_bloc[bloc] = pp

    return pp_by_bloc


def _validate_cambridge_blocs(
    config: BlocSlateConfig,
    majority_bloc: Optional[str] = None,
    minority_bloc: Optional[str] = None,
) -> tuple[str, str]:
    """
    Validates the parameters passed to the Cambridge model and determines the majority and minority
    blocs.

    Args:
        config (BlocSlateConfig): Configuration object containing all necessary parameters for
            working with a bloc-slate ballot generator.
        majority_bloc (Optional[str]): Name of the bloc corresponding to the majority bloc.
            Defaults to whichever bloc has majority via ``bloc_voter_prop``.
        minority_bloc (Optional[str]): Name of the bloc corresponding to the minority bloc.
            Defaults to whichever bloc has minority via ``bloc_voter_prop``.

    Returns:
        tuple[str, str]: A tuple containing the names of the majority and minority blocs.
    """
    if len(config.slates) > 2:
        raise UserWarning(
            f"This model currently only supports at two blocs, but you \
                          passed {len(config.slates)}"
        )

    if (majority_bloc is None) != (minority_bloc is None):
        raise ValueError(
            "Both 'majority_bloc' and 'minority' must be provided or not provided. "
            "You have provided only one."
        )

    elif majority_bloc is not None and majority_bloc == minority_bloc:
        raise ValueError("majority and minority bloc must be distinct.")

    if majority_bloc is None:
        majority_bloc = [
            bloc for bloc, prop in config.bloc_proportions.items() if prop >= 0.5
        ][0]
    else:
        majority_bloc = majority_bloc

    if minority_bloc is None:
        minority_bloc = [
            bloc for bloc in config.bloc_proportions.keys() if bloc != majority_bloc
        ][0]
    else:
        minority_bloc = minority_bloc

    if set(config.blocs) != set(config.slates):
        raise ValueError(
            "This model requires that a bloc and it's preferred slate have the same name. "
            f"You passed blocs {config.blocs} and slates {config.slates}"
        )

    return majority_bloc, minority_bloc


# =================================================
# ================= API Functions =================
# =================================================


def cambridge_profiles_by_bloc_generator(
    config: BlocSlateConfig,
    *,
    path: Optional[Path] = None,
    majority_bloc: Optional[str] = None,
    minority_bloc: Optional[str] = None,
    # historical_majority: Optional[str] = "W",
    # historical_minority: Optional[str] = "C",
    group_ballots: bool = False,
) -> dict[str, RankProfile]:
    """
    Generates a dictionary mapping bloc names to RankProfiles using historical RCV elections occurring
    in Cambridge, MA.

    Alternative election data can be used if specified. The historical data must be contianed
    at the path specified by the 'path' keyword argument, and the data must be a pickle file
    containing a dictionary mapping ballot types with labels 'W' and 'C' (i.e. tuples of the
    form ('W','C','W',...) and the like) to their frequencies. Here 'W' indicates the majority
    bloc and slate and 'C' indicates the minority bloc and slate. Assumes that there are two
    blocs which mimic the formatting of the historical Cambridge data.

    Based on cohesion parameters, decides if a voter casts their top choice within their bloc
    or in the opposing bloc. Then uses historical data; given their first choice, to choose a
    ballot type from the historical distribution.

    Args:
        config (BlocSlateConfig): Configuration object containing all necessary parameters for
            working with a bloc-slate ballot generator.

    Kwargs:
        path (Optional[Path]): File path to an election data file to sample from. If none, will
            default to Cambridge election data that ships with VoteKit
        majority_bloc (Optional[str]): Name of the bloc corresponding to the majority bloc. Defaults to
            whichever bloc has majority via ``bloc_voter_prop``.
        minority_bloc (Optional[str]): Name of the bloc corresponding to the minority bloc. Defaults to
            whichever bloc has minority via ``bloc_voter_prop``.
        group_ballots (bool): If True, groups identical ballots in the resulting profiles.
            Defaults to False.


    Returns:
        dict[str, RankProfile]: A dictionary whose keys are bloc strings and values are
            ``RankProfile`` objects representing the generated preference profiles for each bloc.
    """
    majority_bloc, minority_bloc = _validate_cambridge_blocs(
        config, majority_bloc=majority_bloc, minority_bloc=minority_bloc
    )

    if path is None:
        BASE_DIR = Path(__file__).resolve().parent
        DATA_DIR = BASE_DIR / "data/"
        path = Path(DATA_DIR, "Cambridge_09to17_ballot_types.p")

    pp_by_bloc = _inner_cambridge_sampler(
        config,
        path=path,
        majority_bloc=majority_bloc,
        minority_bloc=minority_bloc,
        historical_majority="W",
        historical_minority="C",
    )

    if group_ballots:
        for bloc, profile in pp_by_bloc.items():
            pp_by_bloc[bloc] = profile.group_ballots()
    return pp_by_bloc


def cambridge_profile_generator(
    config: BlocSlateConfig,
    *,
    path: Optional[Path] = None,
    majority_bloc: Optional[str] = None,
    minority_bloc: Optional[str] = None,
    # historical_majority: Optional[str] = "W",
    # historical_minority: Optional[str] = "C",
    group_ballots: bool = False,
) -> RankProfile:
    """
    Generates a RankProfile using historical RCV elections occurring in Cambridge, MA.

    Alternative election data can be used if specified. The historical data must be contianed
    at the path specified by the 'path' keyword argument, and the data must be a pickle file
    containing a dictionary mapping ballot types with labels 'W' and 'C' (i.e. tuples of the
    form ('W','C','W',...) and the like) to their frequencies. Here 'W' indicates the majority
    bloc and slate and 'C' indicates the minority bloc and slate. Assumes that there are two
    blocs which mimic the formatting of the historical Cambridge data.

    Based on cohesion parameters, decides if a voter casts their top choice within their bloc
    or in the opposing bloc. Then uses historical data; given their first choice, to choose a
    ballot type from the historical distribution.

    Args:
        config (BlocSlateConfig): Configuration object containing all necessary parameters for
            working with a bloc-slate ballot generator.

    Kwargs:
        path (Optional[Path]): File path to an election data file to sample from. If none, will
            default to Cambridge election data that ships with VoteKit
        majority_bloc (Optional[str]): Name of the bloc corresponding to the majority bloc. Defaults to
            whichever bloc has majority via ``bloc_voter_prop``.
        minority_bloc (Optional[str]): Name of the bloc corresponding to the minority bloc. Defaults to
            whichever bloc has minority via ``bloc_voter_prop``.
        group_ballots (bool): If True, groups identical ballots in the resulting profiles.
            Defaults to False.


    Returns:
        RankProfile: A ``RankProfile`` objects representing the joint preference profile over all
            blocs.
    """
    config.is_valid(raise_errors=True)
    majority_bloc, minority_bloc = _validate_cambridge_blocs(
        config, majority_bloc=majority_bloc, minority_bloc=minority_bloc
    )

    if path is None:
        BASE_DIR = Path(__file__).resolve().parent
        DATA_DIR = BASE_DIR / "data/"
        path = Path(DATA_DIR, "Cambridge_09to17_ballot_types.p")

    pp_by_bloc = _inner_cambridge_sampler(
        config,
        path=path,
        majority_bloc=majority_bloc,
        minority_bloc=minority_bloc,
        historical_majority="W",
        historical_minority="C",
    )

    profile = RankProfile()
    for prof in pp_by_bloc.values():
        profile += prof

    if group_ballots:
        profile = profile.group_ballots()

    return profile



import numpy as np
from matplotlib.axes import Axes
from typing import Optional, Dict, List, Tuple, Union
from matplotlib.colors import Colormap, TwoSlopeNorm
import matplotlib.pyplot as plt
import seaborn as sns
import matplotlib.patches as mpatches


def _validate_heatmap_inputs(
    *,
    matrix: np.ndarray,
    row_labels: Optional[List[str]] = None,
    column_labels: Optional[List[str]] = None,
    row_legend: Optional[Dict[str, str]] = None,
    column_legend: Optional[Dict[str, str]] = None,
) -> None:
    """
    Used internally for the `matrix_heatmap` function to validate the inputs to the function.

    Args:
        matrix (np.ndarray): A 2D numpy array containing the data to be plotted.
        row_labels (Optional[List[str]]): A list of strings containing the labels for the rows
            of the heatmap. Defaults to None.
        column_labels (Optional[List[str]]): A list of strings containing the labels for the
            columns of the heatmap. Defaults to None.
        row_legend (Optional[Dict[str, str]]): A dictionary mapping row labels to legend
            descriptions. Defaults to None.
        column_legend (Optional[Dict[str, str]]): A dictionary mapping column labels to legend
            descriptions. Defaults to None.

    Raises:
        ValueError: If the matrix is not a 2D numpy array.
        ValueError: If the row labels are not of the correct length.
        ValueError: If the column labels are not of the correct length.
        ValueError: If the keys of the row legend are do not match the row labels.
        ValueError: If the keys of the column legend are do not match the column labels.
        ValueError: If the legend descriptions in the row legend and column legend are in conflict.
    """
    if matrix.ndim != 2:
        raise ValueError(
            f"Please provide a 2D matrix to plot. Found a {matrix.ndim}-D matrix."
        )

    n_rows = matrix.shape[0]
    n_cols = matrix.shape[1]

    if row_labels is not None:
        if n_rows != len(row_labels):
            raise ValueError(
                f"Please provide {n_rows} labels for the rows of the "
                f"matrix. Found {len(row_labels)} labels."
            )

    if column_labels is not None:
        if n_cols != len(column_labels):
            raise ValueError(
                f"Please provide {n_cols} labels for the columns of the "
                f"matrix. Found {len(column_labels)} labels."
            )

    if row_legend is not None and column_legend is not None:
        for label, description in column_legend.items():
            desc = row_legend.get(label, description)
            if desc != description:
                raise ValueError(
                    f"Conflicting legend descriptions for '{label}': "
                    f"got '{description}' and '{desc}'."
                )

    if row_legend is not None and list(row_legend.keys()) != row_labels:
        raise ValueError("Row labels do not match row legend keys.")

    if column_legend is not None and list(column_legend.keys()) != column_labels:
        raise ValueError("Column labels do not match column legend keys.")


def _add_text_to_heatmap(
    *,
    heatmap: Axes,
    n_decimals_to_display: int,
    matrix: np.ndarray,
    cell_font_size: Optional[int] = None,
) -> Axes:
    """
    Adds the text values to the heatmap cells. This function dynamically determines
    the font size based on the number of cells in the figure, the figure size, and
    the length of the numbers to ensure readability.

    Args:
        heatmap (matplotlib.axes.Axes): The matplotlib axis containing the heatmap.
        n_decimals_to_display (int): The number of decimal places to display for the values
            in the heatmap.
        matrix (np.ndarray): A 2D numpy array containing the data to be plotted.
        cell_font_size (Optional[int]): The base font size to use for the cell values.
            If None, the font size will be dynamically determined.

    Returns:
        matplotlib.axes.Axes: The matplotlib axis containing the heatmap with the cell values
            and text values added.
    """
    nrows, ncols = matrix.shape
    quadmesh = heatmap.collections[0]

    max_chars = max(
        len(f"{val:.{n_decimals_to_display}f}") + (3 if val < 0 else 2)
        for val in matrix.flatten()
    )

    if cell_font_size is not None:
        font_size = cell_font_size
    else:
        fig = heatmap.get_figure()
        if fig:
            fig.canvas.draw()

            # ignoring mypy error, mypy not up to date with matplotlib get_renderer
            renderer = fig.canvas.get_renderer()  # type: ignore[attr-defined]

            # bounding box in display coords
            bbox = heatmap.get_window_extent(renderer=renderer)
            width_pts = bbox.width

            cell_width_pts = width_pts / ncols

            font_size = int(cell_width_pts / max_chars)

    for i in range(nrows):
        for j in range(ncols):
            val = matrix[i, j]
            txt = "N/A" if np.isnan(val) else f"{val:.{n_decimals_to_display}f}"
            txt = "inf" if np.isinf(val) else f"{val:.{n_decimals_to_display}f}"

            # Normalize the cell value between 0 and 1, then get the RGBA color
            norm_val = quadmesh.norm(val)
            r, g, b, _ = quadmesh.cmap(norm_val)

            # Simple brightness measure: average of R, G, B
            brightness = (r + g + b) / 3
            txt_color = "black" if brightness > 0.5 else "white"
            txt_color = "white" if np.isnan(val) else txt_color
            txt_color = "white" if np.isinf(val) else txt_color

            heatmap.text(
                j + 0.5,
                i + 0.5,
                txt,
                ha="center",
                va="center",
                color=txt_color,
                fontsize=font_size,
            )

    return heatmap


def _add_legend_to_heatmap(
    *,
    row_and_col_legends: Dict[str, str],
    ax: Axes,
    legend_font_size: float,
    legend_loc: str,
    legend_bbox_to_anchor: Tuple[float, float],
) -> Axes:
    """
    Adds a legend to a heatmap.

    Args:
        row_and_col_legends (Dict[str, str]): A dictionary mapping row and column labels to
            legend descriptions.
        ax (matplotlib.axes.Axes): The matplotlib axis to add the legend to.
        legend_font_size (float): The font size to use for the legend.
        legend_loc (str): The location to place the legend.
        legend_bbox_to_anchor (Tuple[float, float]): The bounding box to anchor the legend to.

    Returns:
        matplotlib.axes.Axes: The matplotlib axis containing with the updated legend.
    """
    proxy_artists = []
    proxy_labels = []

    if len(row_and_col_legends) != 0:
        for label, description in row_and_col_legends.items():
            patch = mpatches.Patch(color="white", label=f"{label}: {description}")
            proxy_artists.append(patch)
            proxy_labels.append(f"{label}: {description}")

    if proxy_artists:
        leg = ax.legend(
            handles=proxy_artists,
            labels=proxy_labels,
            loc=legend_loc,
            bbox_to_anchor=legend_bbox_to_anchor,
            fontsize=legend_font_size,
            ncol=len(proxy_labels) // 15 + 1,
            frameon=True,
            borderaxespad=0.0,
            handlelength=0,
            handletextpad=0,
            fancybox=True,
        )

        for item in leg.legend_handles:
            if item:
                item.set_visible(False)

    return ax


def matrix_heatmap(
    matrix: np.ndarray,
    *,
    ax: Optional[Axes] = None,
    show_cell_values: bool = True,
    n_decimals_to_display: int = 2,
    row_labels: Optional[List[str]] = None,
    row_label_rotation: Optional[float] = None,
    row_legend: Optional[dict[str, str]] = None,
    column_labels: Optional[List[str]] = None,
    column_label_rotation: Optional[float] = None,
    column_legend: Optional[dict[str, str]] = None,
    cell_color_map: Optional[Union[str, Colormap]] = None,
    cell_font_size: Optional[int] = None,
    cell_spacing: float = 0.5,
    cell_divider_color: str = "white",
    show_colorbar: bool = False,
    legend_font_size: float = 10.0,
    legend_location: str = "center left",
    legend_bbox_to_anchor: Tuple[float, float] = (1.03, 0.5),
) -> Axes:
    """
    Basic function for plotting a matrix as a heatmap.

    Args:
        matrix (np.ndarray): A 2D numpy array containing the data to be plotted.
        ax (matplotlib.axes.Axes, optional): The matplotlib axis to plot on. Defaults to None,
            in which case an axis is created.
        show_cell_values (bool): Whether to show the values of the cells in the heatmap. These
            values are shown in the center of each cell and are dynamically formatted to be
            human-readable.  Defaults to True.
        n_decimals_to_display (int): The number of decimal places to display for the values
            in the heatmap.  Defaults to 2.
        row_labels (Optional(List[str])): A list of strings containing the labels for the rows
            of the heatmap. Defaults to None.
        row_label_rotation (Optional(float)): The rotation to apply to the row labels.
            Defaults to None.
        row_legend (Optional(Dict[str, str])): A dictionary mapping row labels to legend
            descriptions. Defaults to None.
        column_labels (Optional(List[str])): A list of strings containing the labels for the
            columns of the heatmap. Defaults to None.
        column_label_rotation (Optional(float)): The rotation to apply to the column labels.
            Defaults to None.
        column_legend (Optional(Dict[str, str])): A dictionary mapping column labels to legend
            descriptions. Defaults to None.
        cell_color_map (Optional(Union[str, matplotlib.colors.Colormap])): The color map to use
            for the heatmap. Defaults to `PRGn` if the matrix contains negative values and
            `Greens` otherwise.
        cell_font_size (Optional(int)): The font size to use for the cell values. Defaults to
            None, which will then use dynamic font size based on the number of cells and the
            figure size.
        cell_spacing (float): The spacing between the cells in the heatmap. Defaults to 0.5.
        cell_divider_color (str): The color to use for the cell dividers for spacing cells.
            Defaults to "white".
        show_colorbar (bool): Whether to show the colorbar for the heatmap. Defaults to False.
        legend_font_size (float): The font size to use for the legend. Defaults to 10.0.
        legend_location (str): The location to place the legend. Defaults to "center left".
        legend_bbox_to_anchor (Tuple[float, float]): The bounding box to anchor the legend to.
            Defaults to (1.03, 0.5).

    Returns:
        matplotlib.axes.Axes: The matplotlib axis containing the heatmap.
    """

    if ax is None:
        _, ax = plt.subplots()

    _validate_heatmap_inputs(
        matrix=matrix,
        row_labels=row_labels,
        column_labels=column_labels,
        row_legend=row_legend,
        column_legend=column_legend,
    )

    row_and_col_legends = dict()
    if row_legend is not None:
        row_and_col_legends.update(row_legend)
    if column_legend is not None:
        row_and_col_legends.update(column_legend)

    ax.xaxis.set_ticks_position("top")
    ax.yaxis.set_ticks_position("left")

    vmin = np.nanmin(matrix)
    vmax = np.nanmax(matrix)
    norm: Union[TwoSlopeNorm, str, None] = (
        TwoSlopeNorm(vmin=vmin, vcenter=0.0, vmax=vmax) if (vmin < 0 < vmax) else None
    )

    if cell_color_map is None:
        if np.nanmin(matrix) < 0:
            cell_color_map = sns.color_palette("PRGn", as_cmap=True)
            norm = TwoSlopeNorm(
                vmin=np.nanmin(matrix), vcenter=0.0, vmax=np.nanmax(matrix)
            )
        else:
            cell_color_map = sns.color_palette("Greens", as_cmap=True)
            norm = "linear"

    heatmap = sns.heatmap(
        matrix,
        ax=ax,
        cmap=cell_color_map,
        norm=norm,
        fmt=f".{n_decimals_to_display}f",
        linewidths=cell_spacing,
        linecolor=cell_divider_color,
        cbar=show_colorbar,
        yticklabels=row_labels if row_labels is not None else False,
        xticklabels=column_labels if column_labels is not None else False,
    )

    plt.gca().set_facecolor("black")  # inf cells

    if show_cell_values:
        heatmap = _add_text_to_heatmap(
            heatmap=heatmap,
            n_decimals_to_display=n_decimals_to_display,
            matrix=matrix,
            cell_font_size=cell_font_size,
        )

    if len(row_and_col_legends) > 0:
        heatmap = _add_legend_to_heatmap(
            row_and_col_legends=row_and_col_legends,
            ax=heatmap,
            legend_font_size=legend_font_size,
            legend_loc=legend_location,
            legend_bbox_to_anchor=legend_bbox_to_anchor,
        )

    if column_label_rotation is not None:
        ax.set_xticklabels(
            ax.get_xticklabels(),
            rotation=column_label_rotation,
            ha="left",
            rotation_mode="anchor",
        )
    if row_label_rotation is not None:
        ax.set_yticklabels(
            ax.get_yticklabels(), rotation=row_label_rotation, rotation_mode="anchor"
        )

    return ax



from votekit.elections.election_types.ranking.abstract_ranking import RankingElection
from votekit.pref_profile import RankProfile
from votekit.ballot import RankBallot
from votekit.elections.election_state import ElectionState
from votekit.utils import (
    first_place_votes,
    score_dict_to_ranking,
    tiebreak_set,
)
from votekit.cleaning import remove_and_condense_rank_profile
import numpy as np
from typing import Optional, Literal
from functools import partial


class PluralityVeto(RankingElection):
    """
    Scores each candidate by their plurality (number of first) place votes,
    then in a randomized order it lets each voter decrement the score of their
    least favorite candidate. The candidate with the largest score at the end is
    then chosen as the winner.

    Args:
      profile (RankProfile): RankProfile to run election on. Note that
        ballots must have integer weights to be considered valid for this mechanism.
      m (int): Number of seats to elect.
      tiebreak (str, optional): Tiebreak method to use. Options are None, 'random', and 'borda'.
            Defaults to None, in which case a tie raises a ValueError.
      fpv_tie_convention (Literal["high", "average", "low"], optional): How to award points
            for tied first place votes. Defaults to "average", where if n candidates are tied for
            first, each receives 1/n points. "high" would award them each one point, and "low" 0.
            Only used by ``score_function`` parameter.

    Attributes:
        m (int): The number of seats to be filled in the election.
        tiebreak (Optional[str]): The tiebreak method selected for the election. Could be None,
            'random', or 'borda'.
        ballot_list (List[Ballot]): A list of Ballot instances representing the decondensed ballots
            where each ballot has a weight of 1.
        random_order (List[int]): A list of indices representing the randomized order in which
            voters' preferences are processed.
        preference_index (List[int]): A list of integers where each entry corresponds to the
            index of the least preferred candidate for each ballot.
        eliminated_dict (Dict[str, bool]): A dictionary mapping each candidate to a boolean value
            indicating whether the candidate has been eliminated.

    Raises:
        ValueError: If ``m`` is not positive or if it exceeds the number of candidates.
    """

    def __init__(
        self,
        profile: RankProfile,
        m: int,
        tiebreak: Optional[str] = None,
        fpv_tie_convention: Literal["high", "low", "average"] = "average",
    ):
        """
        Initializes the Plurality Veto election class.

        This constructor sets up the initial state of the election, including
        validating the profile, decondensing ballots (necessary for the running the steps of the
        election), and preparing other necessary data structures.

        Args:
            profile (RankProfile): The preference profile to be used in the election.
            m (int): The number of seats to be filled in the election.
            tiebreak (Optional[str]): The method used to resolve ties. Defaults to None.

        Raises:
            ValueError: If ``m`` is less than or equal to 0, or if ``m`` exceeds the number of
                candidates in the profile.
            AttributeError: If there are any ballots with ties and no tiebreak method is specified.
            TypeError: If the profile contains invalid ballots.
        """
        self._pv_validate_profile(profile)

        if m <= 0:
            raise ValueError("m must be positive.")
        elif len(profile.candidates_cast) < m:
            raise ValueError("Not enough candidates received votes to be elected.")

        self.m = m
        self.tiebreak = tiebreak

        if self.tiebreak is None and profile.ballots is not None:
            for ballot in profile.ballots:
                if ballot.ranking is not None and any(
                    len(s) > 1 for s in ballot.ranking
                ):
                    raise AttributeError(
                        "Found Ballots with ties but no tiebreak method was specified."
                    )

        # Ungroup ballots
        ballots = profile.ballots
        new_ballots = [RankBallot() for _ in range(int(profile.total_ballot_wt))]
        bidx = 0
        for b in ballots:
            for _ in range(int(b.weight)):
                new_ballots[bidx] = RankBallot(
                    ranking=b.ranking,
                    weight=1,
                    voter_set=b.voter_set,
                )
                bidx += 1

        profile = RankProfile(ballots=tuple(new_ballots), candidates=profile.candidates)

        self.ballot_list = profile.ballots
        self.random_order = list(range(int(profile.num_ballots)))
        np.random.shuffle(self.random_order)
        self.preference_index = [
            len(ballot.ranking) - 1 if ballot.ranking else -1
            for ballot in self.ballot_list
        ]

        self.eliminated_dict = {c: False for c in profile.candidates}

        super().__init__(
            profile,
            score_function=partial(
                first_place_votes, tie_convention=fpv_tie_convention
            ),
        )

    def _pv_validate_profile(self, profile: RankProfile):
        """
        Validate that each ballot has a ranking and that each
        ballot has integer weight.
        """
        for ballot in profile.ballots:
            if ballot.ranking is None:
                raise TypeError("Ballots must have rankings.")
            elif int(ballot.weight) != ballot.weight:
                raise TypeError(f"Ballot {ballot} has non-integer weight.")

    def _is_finished(self) -> bool:
        cands_elected = [len(s) for s in self.get_elected()]
        return sum(cands_elected) >= self.m

    def _run_step(
        self, profile: RankProfile, prev_state: ElectionState, store_states=False
    ) -> RankProfile:
        """
        Run one step of an election from the given profile and previous state.
        If m candidates have not yet been elected:
        if the current round is 0, count plurality scores and remove all
        candidates with a score of 0 to eliminated. Otherwise, this method
        eliminates a single candidate by decrementing scores of candidates on
        the current ticket until someone falls to a score of 0. Once we find
        that there are exactly m candidates remaining on the ticket, elect all
        of them.

        Args:
            profile (RankProfile): Profile of ballots.
            prev_state (ElectionState): The previous ElectionState.
            store_states (bool, optional): True if `self.election_states` should be updated with the
                ElectionState generated by this round. This should only be True when used by
                `self._run_election()`. Defaults to False.

        Returns:
            RankProfile: The profile of ballots after the round is completed.
        """
        remaining_count = len(self.eliminated_dict) - sum(self.eliminated_dict.values())

        if remaining_count == self.m:
            # move all to elected, this is the last round
            elected = prev_state.remaining
            new_profile = RankProfile()

            if store_states:
                if self.score_function:
                    scores = self.score_function(new_profile)
                new_state = ElectionState(
                    round_number=prev_state.round_number + 1,
                    elected=elected,
                    scores=scores,
                )

            self.election_states.append(new_state)

        else:
            tiebreaks = {}
            new_scores = {c: s for c, s in prev_state.scores.items()}

            eliminated_cands = []
            if prev_state.round_number == 0:
                eliminated_cands = [
                    c for c, score in prev_state.scores.items() if score <= 0
                ]

            for rand_index, ballot_index in enumerate(self.random_order):
                if not self.preference_index[ballot_index] < 0:
                    ballot = self.ballot_list[ballot_index]
                    last_place = self.preference_index[ballot_index]

                    if ballot.ranking:
                        if len(ballot.ranking[last_place]) > 1:
                            if self.tiebreak:
                                tiebroken_ranking = tiebreak_set(
                                    ballot.ranking[last_place], profile, self.tiebreak
                                )
                            tiebreaks = {ballot.ranking[last_place]: tiebroken_ranking}
                        else:
                            tiebroken_ranking = (ballot.ranking[last_place],)

                    least_preferred = list(tiebroken_ranking[-1])[0]
                    new_scores[least_preferred] -= 1

                    if new_scores[least_preferred] <= 0:
                        eliminated_cands.append(least_preferred)
                        break

            # Circularly shift the randomized order array so that
            # we can continue where we left off in the next round
            self.random_order = (
                self.random_order[rand_index + 1 :]
                + self.random_order[: rand_index + 1]
            )

            for c in eliminated_cands:
                self.eliminated_dict[c] = True

            new_profile = remove_and_condense_rank_profile(
                eliminated_cands,
                profile,
                remove_zero_weight_ballots=False,
                remove_empty_ballots=False,
            )

            self.ballot_list = new_profile.ballots

            self.preference_index = [
                len(ballot.ranking) - 1 if ballot.ranking else -1
                for ballot in self.ballot_list
            ]

            if store_states:
                eliminated = (frozenset(eliminated_cands),)

                score_ballots = tuple(
                    ballot for ballot in new_profile.ballots if ballot.ranking
                )
                score_profile = RankProfile(ballots=score_ballots)

                if self.score_function:
                    scores = self.score_function(score_profile)

                remaining = score_dict_to_ranking(scores)
                new_state = ElectionState(
                    round_number=prev_state.round_number + 1,
                    eliminated=eliminated,
                    remaining=remaining,
                    scores=scores,
                    tiebreaks=tiebreaks,
                )

                self.election_states.append(new_state)
        return new_profile



from votekit.elections.election_types.ranking.abstract_ranking import RankingElection
from votekit.pref_profile import RankProfile
from votekit.elections.election_state import ElectionState
from votekit.cleaning import remove_and_condense_rank_profile
from votekit.utils import elect_cands_from_set_ranking, borda_scores
from votekit.graphs import PairwiseComparisonGraph


class CondoBorda(RankingElection):
    """
    Just like DominatingSets (Smith method), but user gets to choose the number of winners,
    :math:`m`.  Ties are broken with Borda scores.

    Args:
        profile (RankProfile): Profile to conduct election on.
        m (int, optional): Number of seats to elect. Defaults to 1.

    """

    def __init__(self, profile: RankProfile, m: int = 1):
        if len(profile.candidates_cast) < m:
            raise ValueError("Not enough candidates received votes to be elected.")
        self.m = m
        super().__init__(profile, score_function=borda_scores)

    def _is_finished(self):
        # single round election
        if len(self.election_states) == 2:
            return True
        return False

    def _run_step(
        self, profile: RankProfile, prev_state: ElectionState, store_states=False
    ) -> RankProfile:
        """
        Run one step of an election from the given profile and previous state.
        Compute the dominating tiers, and return the top :math:`m` candidates by tier.
        Breaks ties using Borda scores.

        Args:
            profile (RankProfile): Profile of ballots.
            prev_state (ElectionState): The previous ElectionState.
            store_states (bool, optional): True if `self.election_states` should be updated with the
                ElectionState generated by this round. This should only be True when used by
                `self._run_election()`. Defaults to False.

        Returns:
            RankProfile: The profile of ballots after the round is completed.
        """
        pwc_graph = PairwiseComparisonGraph(profile)
        dominating_tiers = pwc_graph.get_dominating_tiers()

        dt_ranking = tuple([frozenset(s) for s in dominating_tiers])
        elected, remaining, tiebreak = elect_cands_from_set_ranking(
            dt_ranking, self.m, profile, tiebreak="borda"
        )

        if tiebreak:
            tiebreaks = {tiebreak[0]: tiebreak[1]}
        else:
            tiebreaks = {}

        new_profile = remove_and_condense_rank_profile(
            [c for s in elected for c in s], profile
        )

        if store_states:
            self.election_states.append(
                ElectionState(
                    round_number=prev_state.round_number + 1,
                    elected=elected,
                    remaining=remaining,
                    tiebreaks=tiebreaks,
                    scores=borda_scores(new_profile),
                )
            )

        return new_profile



from __future__ import annotations
from votekit.pref_profile import RankProfile, ScoreProfile
import warnings
import pandas as pd


class CleanedRankProfile(RankProfile):
    """
    CleanedRankProfile class, which is used to keep track of how ballots are altered from the
    original
    profile. In addition to a custom __str__ method, this class implements a collection of sets
    that track the indices of the ballot dataframe, and how they are changed by different cleaning
    rules. It also retains the parent profile of the CleanedRankProfile, allowing for full recovery
    of the cleaning steps.

    Args:
        ballots (tuple[Ballot], optional): Tuple of ``Ballot`` objects. Defaults to empty tuple.
        candidates (tuple[str], optional): Tuple of candidate strings. Defaults to empty tuple.
            If empty, computes this from any candidate listed on a ballot with positive weight.
        max_ranking_length (int, optional): The length of the longest allowable ballot, i.e., how
            many candidates are allowed to be ranked in an election. Defaults to longest observed
            ballot.
        parent_profile (PreferenceProfile | CleanedProfile): The profile that was altered.
            If you apply multiple cleaning functions, the parent is always the profile immediately
            before cleaning, so you need to recurse to get the original, uncleaned profile.
        df_index_column (list[int]): The indices of the ballots in the df from the parent profile.
        no_wt_altr_idxs (set[int], optional): Set of indices of ballots that have
            0 weight as a result of cleaning. Indices are with respect
            to ``parent_profile.df``.
        no_rank_altr_idxs (set[int], optional): Set of indices of
            ballots that have no ranking as a result of cleaning. Indices are with
            respect to ``parent_profile.df``.
        nonempty_altr_idxs (set[int], optional):  Set of indices of ballots that
            have been altered but still have weight and (ranking or score) as a result of cleaning.
            Indices are with respect to ``parent_profile.df``.
        unaltr_idxs (set[int], optional):  Set of indices of ballots that have
            been unaltered by cleaning. Indices are with respect to ``parent_profile.df``.


    """

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        return instance

    def __init__(
        self,
        parent_profile: RankProfile | CleanedRankProfile = (RankProfile()),
        df_index_column: list[int] = [],
        no_wt_altr_idxs: set[int] = set(),
        no_rank_altr_idxs: set[int] = set(),
        nonempty_altr_idxs: set[int] = set(),
        unaltr_idxs: set[int] = set(),
        **kwargs,
    ):

        self.parent_profile = parent_profile
        self.df_index_column = df_index_column
        self.no_wt_altr_idxs = no_wt_altr_idxs
        self.no_rank_altr_idxs = no_rank_altr_idxs
        self.nonempty_altr_idxs = nonempty_altr_idxs
        self.unaltr_idxs = unaltr_idxs

        self._indices_must_match_parent_df()
        super().__init__(**kwargs)
        self._reindex_ballot_df()

    def _indices_must_match_parent_df(self) -> None:
        """
        Validate the index sets.

        Raises:
            ValueError: index set is not a subset of the parent df index.
            ValueError: union of indices is not equal to the parent df index.

        """
        index_sets = [
            self.no_wt_altr_idxs,
            self.no_rank_altr_idxs,
            self.nonempty_altr_idxs,
            self.unaltr_idxs,
        ]

        if not self.no_wt_altr_idxs.issubset(self.parent_profile.df.index):
            set_minus = self.no_wt_altr_idxs.difference(self.parent_profile.df.index)
            raise ValueError(
                (
                    "no_wt_altr_idxs is not a subset of the"
                    " parent profile df index column. Here are the indices found in no_wt_altr_idxs"
                    f" but not the parent profile df index column: {set_minus}"
                )
            )

        if not self.no_rank_altr_idxs.issubset(self.parent_profile.df.index):
            set_minus = self.no_rank_altr_idxs.difference(self.parent_profile.df.index)
            raise ValueError(
                (
                    "no_rank_altr_idxs is not a subset of the"
                    " parent profile df index column. Here are the indices found in"
                    " no_rank_altr_idxs but not the parent profile df index column: "
                    f"{set_minus}"
                )
            )

        if not self.nonempty_altr_idxs.issubset(self.parent_profile.df.index):
            set_minus = self.nonempty_altr_idxs.difference(self.parent_profile.df.index)
            raise ValueError(
                (
                    "nonempty_altr_idxs is not a subset of the"
                    " parent profile df index column. Here are the indices found in"
                    " nonempty_altr_idxs but not the parent profile df index column: "
                    f"{set_minus}"
                )
            )

        if not self.unaltr_idxs.issubset(self.parent_profile.df.index):
            set_minus = self.unaltr_idxs.difference(self.parent_profile.df.index)
            raise ValueError(
                (
                    "unaltr_idxs is not a subset of the"
                    " parent profile df index column. Here are the indices found in"
                    " unaltr_idxs but not the parent profile df index column: "
                    f"{set_minus}"
                )
            )

        if set().union(*index_sets) != set(self.parent_profile.df.index):
            sym_dif = (
                set()
                .union(*index_sets)
                .symmetric_difference(self.parent_profile.df.index)
            )
            raise ValueError(
                (
                    "Union of ballot indices must equal the parent profile df index "
                    f"column. Here are the indices in one but not the other: {sym_dif}"
                )
            )

    def _reindex_ballot_df(self) -> None:
        """
        Reindex the df to keep track of which ballots are which.

        Raises:
            ValueError: If df_index_column does not have the same length as the ballot list.

        """
        if len(self.df) != len(self.df_index_column):
            raise ValueError(
                "df_index_column does not have the same length as the ballot list."
            )

        df_copy = self.df.copy()
        df_copy.index = pd.Index(self.df_index_column)
        df_copy.index.name = "Ballot Index"
        object.__setattr__(self, "df", df_copy)

    def group_ballots(self) -> RankProfile:
        warnings.warn(
            (
                "Grouping the ballots of a CleanedRankProfile will return a RankProfile"
                " since this operation resets ballot indices."
            )
        )
        return super().group_ballots()

    def __str__(self) -> str:

        return "Profile has been cleaned\n" + super().__str__()

    __repr__ = __str__


class CleanedScoreProfile(ScoreProfile):
    """
    CleanedScoreProfile class, which is used to keep track of how ballots are altered from the
    original
    profile. In addition to a custom __str__ method, this class implements a collection of sets
    that track the indices of the ballot dataframe, and how they are changed by different cleaning
    rules. It also retains the parent profile of the CleanedScoreProfile, allowing for full recovery
    of the cleaning steps.

    Args:
        ballots (tuple[Ballot], optional): Tuple of ``Ballot`` objects. Defaults to empty tuple.
        candidates (tuple[str], optional): Tuple of candidate strings. Defaults to empty tuple.
            If empty, computes this from any candidate listed on a ballot with positive weight.
        parent_profile (ScoreProfile | CleanedScoreProfile): The profile that was altered.
            If you apply multiple cleaning functions, the parent is always the profile immediately
            before cleaning, so you need to recurse to get the original, uncleaned profile.
        df_index_column (list[int]): The indices of the ballots in the df from the parent profile.
        no_wt_altr_idxs (set[int], optional): Set of indices of ballots that have
            0 weight as a result of cleaning. Indices are with respect
            to ``parent_profile.df``.
        no_scores_altr_idxs (set[int], optional): Set of indices of
            ballots that have no scores as a result of cleaning. Indices are with
            respect to ``parent_profile.df``.
        nonempty_altr_idxs (set[int], optional):  Set of indices of ballots that
            have been altered but still have weight and (ranking or score) as a result of cleaning.
            Indices are with respect to ``parent_profile.df``.
        unaltr_idxs (set[int], optional):  Set of indices of ballots that have
            been unaltered by cleaning. Indices are with respect to ``parent_profile.df``.


    """

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        return instance

    def __init__(
        self,
        parent_profile: ScoreProfile | CleanedRankProfile = (ScoreProfile()),
        df_index_column: list[int] = [],
        no_wt_altr_idxs: set[int] = set(),
        no_scores_altr_idxs: set[int] = set(),
        nonempty_altr_idxs: set[int] = set(),
        unaltr_idxs: set[int] = set(),
        **kwargs,
    ):

        self.parent_profile = parent_profile
        self.df_index_column = df_index_column
        self.no_wt_altr_idxs = no_wt_altr_idxs
        self.no_scores_altr_idxs = no_scores_altr_idxs
        self.nonempty_altr_idxs = nonempty_altr_idxs
        self.unaltr_idxs = unaltr_idxs

        self._indices_must_match_parent_df()
        super().__init__(**kwargs)
        self._reindex_ballot_df()

    def _indices_must_match_parent_df(self) -> None:
        """
        Validate the index sets.

        Raises:
            ValueError: index set is not a subset of the parent df index.
            ValueError: union of indices is not equal to the parent df index.

        """
        index_sets = [
            self.no_wt_altr_idxs,
            self.no_scores_altr_idxs,
            self.nonempty_altr_idxs,
            self.unaltr_idxs,
        ]

        if not self.no_wt_altr_idxs.issubset(self.parent_profile.df.index):
            set_minus = self.no_wt_altr_idxs.difference(self.parent_profile.df.index)
            raise ValueError(
                (
                    "no_wt_altr_idxs is not a subset of the"
                    " parent profile df index column. Here are the indices found in no_wt_altr_idxs"
                    f" but not the parent profile df index column: {set_minus}"
                )
            )

        if not self.no_scores_altr_idxs.issubset(self.parent_profile.df.index):
            set_minus = self.no_scores_altr_idxs.difference(
                self.parent_profile.df.index
            )
            raise ValueError(
                (
                    "no_scores_altr_idxs is not a subset of the"
                    " parent profile df index column. Here are the indices found in"
                    " no_scores_altr_idxs but not the parent profile df index column: "
                    f"{set_minus}"
                )
            )

        if not self.nonempty_altr_idxs.issubset(self.parent_profile.df.index):
            set_minus = self.nonempty_altr_idxs.difference(self.parent_profile.df.index)
            raise ValueError(
                (
                    "nonempty_altr_idxs is not a subset of the"
                    " parent profile df index column. Here are the indices found in"
                    " nonempty_altr_idxs but not the parent profile df index column: "
                    f"{set_minus}"
                )
            )

        if not self.unaltr_idxs.issubset(self.parent_profile.df.index):
            set_minus = self.unaltr_idxs.difference(self.parent_profile.df.index)
            raise ValueError(
                (
                    "unaltr_idxs is not a subset of the"
                    " parent profile df index column. Here are the indices found in"
                    " unaltr_idxs but not the parent profile df index column: "
                    f"{set_minus}"
                )
            )

        if set().union(*index_sets) != set(self.parent_profile.df.index):
            sym_dif = (
                set()
                .union(*index_sets)
                .symmetric_difference(self.parent_profile.df.index)
            )
            raise ValueError(
                (
                    "Union of ballot indices must equal the parent profile df index "
                    f"column. Here are the indices in one but not the other: {sym_dif}"
                )
            )

    def _reindex_ballot_df(self) -> None:
        """
        Reindex the df to keep track of which ballots are which.

        Raises:
            ValueError: If df_index_column does not have the same length as the ballot list.

        """
        if len(self.df) != len(self.df_index_column):
            raise ValueError(
                "df_index_column does not have the same length as the ballot list."
            )

        df_copy = self.df.copy()
        df_copy.index = pd.Index(self.df_index_column)
        df_copy.index.name = "Ballot Index"
        object.__setattr__(self, "df", df_copy)

    def group_ballots(self) -> ScoreProfile:
        warnings.warn(
            (
                "Grouping the ballots of a CleanedScoreProfile will return a ScoreProfile"
                " since this operation resets ballot indices."
            )
        )
        return super().group_ballots()

    def __str__(self) -> str:

        return "Profile has been cleaned\n" + super().__str__()

    __repr__ = __str__



from typing import Callable, Optional, Union, Any
from votekit.pref_profile import RankProfile, ScoreProfile
from matplotlib.axes import Axes
from votekit.utils import (
    first_place_votes,
    borda_scores,
    mentions,
    ballot_lengths,
    COLOR_LIST,
)
from functools import partial
from votekit.plots.bar_plot import add_null_keys, multi_bar_plot


def _create_data_dict(
    profile_dict: dict[str, RankProfile | ScoreProfile],
    stat_function: Callable[[RankProfile | ScoreProfile], dict[str, float]],
) -> dict[str, dict[str, float]]:
    """
    Create the correctly formatted dict to pass to ``multi_bar_plot``. Ensures each
    subdictionary has the same keys, and uses the default value 0 if a key is missing.

    Args:
        profile_dict (dict[str, RankProfile | ScoreProfile]): Keys are profile labels and values are
            profiles to plot statistics for.
        stat_function (Callable[[RankProfile | ScoreProfile], dict[str, float]]): Which stat
            to use for the bar plot. Must be a callable that takes a profile and returns
            a dict with str keys and float values.

    Returns:
        dict[str, dict[str, float]]: Data dictionary for ``multi_bar_plot``.

    """

    data_dict = {
        label: stat_function(profile) for label, profile in profile_dict.items()
    }

    return add_null_keys(data_dict)


def multi_profile_bar_plot(
    profile_dict: dict[str, RankProfile | ScoreProfile],
    stat_function: Callable[[RankProfile | ScoreProfile], dict[str, float]],
    normalize: bool = False,
    profile_colors: Optional[dict[str, str]] = None,
    bar_width: Optional[float] = None,
    category_ordering: Optional[list[str]] = None,
    x_axis_name: Optional[str] = None,
    y_axis_name: Optional[str] = None,
    title: Optional[str] = None,
    show_profile_legend: bool = False,
    categories_legend: Optional[dict[str, str]] = None,
    threshold_values: Optional[Union[list[float], float]] = None,
    threshold_kwds: Optional[Union[list[dict], dict]] = None,
    legend_font_size: Optional[float] = None,
    ax: Optional[Axes] = None,
) -> Axes:
    """
    Plot a multi bar plot over a set of preference profiles and some statistic about the profile,
    like ballot length, first place votes for candidate, etc.

    Args:
        profile_dict (dict[str, RankProfile | ScoreProfile]): Keys are profile labels and values are
            profiles to plot statistics for.
        stat_function (Callable[[RankProfile | ScoreProfile], dict[str, float]]): Which stat
            to use for the bar plot. Must be a callable that takes a profile and returns
            a dict with str keys and float values.
        normalize (bool, optional): Whether or not to normalize data. Defaults to False.
        profile_colors (dict[str, str], optional): Dictionary mapping profile labels
            to colors. Defaults to None, in which case we use a subset of ``COLOR_LIST``
            from ``utils`` module. Dictionary keys can be a subset of the profiles.
        bar_width (float, optional): Width of bars. Defaults to None which computes the bar width
            as 0.7 divided by the number of data sets. Must be in the interval :math:`(0,1]`.
        category_ordering (list[str], optional): Ordering of x-labels. Defaults to order retrieved
            from data dictionary.
        x_axis_name (str, optional): Name of x-axis. Defaults to None, which does not plot a name.
        y_axis_name (str, optional): Name of y-axis. Defaults to None, which does not plot a name.
        title (str, optional): Title for the figure. Defaults to None, which does not plot a title.
        show_profile_legend (bool, optional): Whether or not to plot the profile legend.
            Defaults to False. Is automatically shown if any threshold lines have the keyword
            "label" passed through ``threshold_kwds``.
        categories_legend (dict[str, str], optional): Dictionary mapping data categories
            to description. Defaults to None. If provided, generates a second legend for data
            categories.
        threshold_values (Union[list[float], float], optional): List of values to plot horizontal
            lines at. Can be provided as a list or a single float.
        threshold_kwds (Union[list[dict], dict], optional): List of plotting
            keywords for the horizontal lines. Can be a list or single dictionary. These will be
            passed to plt.axhline(). Common keywords include "linestyle", "linewidth", and "label".
            If "label" is passed, automatically plots the data set legend with the labels.
        legend_font_size (float, optional): The font size to use for the legend. Defaults to 10.0
            + the number of categories.
        legend_loc (str, optional): The location parameter to pass to ``Axes.legend(loc=)``.
            Defaults to "center left".
        legend_bbox_to_anchor (Tuple[float, float], otptional): The bounding box to anchor
            the legend to. Defaults to (1, 0.5).
        ax (Axes, optional): A matplotlib axes object to plot the figure on. Defaults to None, in
            which case the function creates and returns a new axes. The figure height is 6 inches
            and the figure width is 3 inches times the number of categories.

    Returns:
        Axes: A ``matplotlib`` axes with a bar plot of the given data.
    """
    data_dict = _create_data_dict(profile_dict, stat_function)

    try:
        ax = multi_bar_plot(
            data_dict,
            normalize=normalize,
            data_set_colors=profile_colors,
            bar_width=bar_width,
            category_ordering=category_ordering,
            x_axis_name=x_axis_name,
            y_axis_name=y_axis_name,
            title=title,
            show_data_set_legend=show_profile_legend,
            categories_legend=categories_legend,
            threshold_values=threshold_values,
            threshold_kwds=threshold_kwds,
            legend_font_size=legend_font_size,
            ax=ax,
        )

        return ax
    except Exception as e:
        if "Cannot plot more than" in str(e):
            raise ValueError(f"Cannot plot more than {len(COLOR_LIST)} profiles.")

        elif str(e) == "category_ordering must be the same length as sub-dictionaries.":
            raise ValueError(
                (
                    "category_ordering must be the same length as the dictionary "
                    f"produced by stat_function: {len(next(iter(data_dict.values())))}."
                )
            )

        else:
            raise e


def multi_profile_borda_plot(
    profile_dict: dict[str, RankProfile],
    borda_kwds: Optional[dict[str, Any]] = None,
    normalize: bool = False,
    profile_colors: Optional[dict[str, str]] = None,
    bar_width: Optional[float] = None,
    candidate_ordering: Optional[list[str]] = None,
    x_axis_name: Optional[str] = None,
    y_axis_name: Optional[str] = None,
    title: Optional[str] = None,
    show_profile_legend: bool = False,
    candidate_legend: Optional[dict[str, str]] = None,
    relabel_candidates_with_int: bool = False,
    threshold_values: Optional[Union[list[float], float]] = None,
    threshold_kwds: Optional[Union[list[dict], dict]] = None,
    legend_font_size: Optional[float] = None,
    ax: Optional[Axes] = None,
) -> Axes:
    """
    Plot the borda scores for a collection of profiles. Wrapper for ``multi_profile_bar_plot``.

    Args:
        profile_dict (dict[str, RankProfile]): Keys are profile labels and values are
            profiles to plot statistics for.
        borda_kwds (dict[str, Any], optional): Keyword arguments to pass to
            ``borda_scores``. Defaults to None, in which case default values for ``borda_scores``
            are used.
        normalize (bool, optional): Whether or not to normalize data. Defaults to False.
        profile_colors (dict[str, str], optional): Dictionary mapping profile labels
            to colors. Defaults to None, in which case we use a subset of ``COLOR_LIST``
            from ``utils`` module. Dictionary keys can be a subset of the profiles.
        bar_width (float, optional): Width of bars. Defaults to None which computes the bar width
            as 0.7 divided by the number of data sets. Must be in the interval :math:`(0,1]`.
        candidate_ordering (list[str], optional): Ordering of x-labels. Defaults to decreasing
            borda scores from the first profile.
        x_axis_name (str, optional): Name of x-axis. Defaults to None, which does not plot a name.
        y_axis_name (str, optional): Name of y-axis. Defaults to None, which does not plot a name.
        title (str, optional): Title for the figure. Defaults to None, which does not plot a title.
        show_profile_legend (bool, optional): Whether or not to plot the profile legend.
            Defaults to False. Is automatically shown if any threshold lines have the keyword
            "label" passed through ``threshold_kwds``.
        candidate_legend (dict[str, str], optional): Dictionary mapping candidates
            to relableing. Defaults to None. If provided, generates a second legend for data
            categories.
        relabel_candidates_with_int (bool, optional): Relabel the candidates with integer labels.
            Defaults to False. If ``candidate_legend`` is passed, those labels supercede.
        threshold_values (Union[list[float], float], optional): List of values to plot horizontal
            lines at. Can be provided as a list or a single float.
        threshold_kwds (Union[list[dict], dict], optional): List of plotting
            keywords for the horizontal lines. Can be a list or single dictionary. These will be
            passed to plt.axhline(). Common keywords include "linestyle", "linewidth", and "label".
            If "label" is passed, automatically plots the data set legend with the labels.
        legend_font_size (float, optional): The font size to use for the legend. Defaults to 10.0
            + the number of categories.
        legend_loc (str, optional): The location parameter to pass to ``Axes.legend(loc=)``.
            Defaults to "center left".
        legend_bbox_to_anchor (Tuple[float, float], otptional): The bounding box to anchor
            the legend to. Defaults to (1, 0.5).
        ax (Axes, optional): A matplotlib axes object to plot the figure on. Defaults to None, in
            which case the function creates and returns a new axes. The figure height is 6 inches
            and the figure width is 3 inches times the number of categories.

    Returns:
        Axes: A ``matplotlib`` axes with a bar plot of the given data.
    """

    stat_function = partial(borda_scores, **borda_kwds) if borda_kwds else borda_scores
    data_dict = _create_data_dict(profile_dict, stat_function)  # type: ignore[arg-type]

    if candidate_ordering is None:
        candidate_ordering = sorted(
            next(iter(data_dict.values())).keys(),
            reverse=True,
            key=lambda x: next(iter(data_dict.values()))[x],
        )

    if relabel_candidates_with_int and candidate_legend is None:
        candidate_legend = {c: str(i + 1) for i, c in enumerate(candidate_ordering)}

    try:
        ax = multi_bar_plot(
            data_dict,
            normalize=normalize,
            data_set_colors=profile_colors,
            bar_width=bar_width,
            category_ordering=candidate_ordering,
            x_axis_name=x_axis_name,
            y_axis_name=y_axis_name,
            title=title,
            show_data_set_legend=show_profile_legend,
            categories_legend=candidate_legend,
            threshold_values=threshold_values,
            threshold_kwds=threshold_kwds,
            legend_font_size=legend_font_size,
            ax=ax,
        )

        return ax
    except Exception as e:
        if "Cannot plot more than" in str(e):
            raise ValueError(f"Cannot plot more than {len(COLOR_LIST)} profiles.")

        elif str(e) == "category_ordering must be the same length as sub-dictionaries.":
            raise ValueError(
                (
                    "candidate_ordering must be the same length as the dictionary "
                    f"produced by borda_scores: {len(next(iter(data_dict.values())))}."
                )
            )

        else:
            raise e


def multi_profile_mentions_plot(
    profile_dict: dict[str, RankProfile],
    mentions_kwds: Optional[dict[str, Any]] = None,
    normalize: bool = False,
    profile_colors: Optional[dict[str, str]] = None,
    bar_width: Optional[float] = None,
    candidate_ordering: Optional[list[str]] = None,
    x_axis_name: Optional[str] = None,
    y_axis_name: Optional[str] = None,
    title: Optional[str] = None,
    show_profile_legend: bool = False,
    candidate_legend: Optional[dict[str, str]] = None,
    relabel_candidates_with_int: bool = False,
    threshold_values: Optional[Union[list[float], float]] = None,
    threshold_kwds: Optional[Union[list[dict], dict]] = None,
    legend_font_size: Optional[float] = None,
    ax: Optional[Axes] = None,
) -> Axes:
    """
    Plot the mentions for a collection of profiles. Wrapper for ``multi_profile_bar_plot``.

    Args:
        profile_dict (dict[str, RankProfile]): Keys are profile labels and values are
            profiles to plot statistics for.
        mentions_kwds (dict[str, Any], optional): Keyword arguments to pass to
            ``mentions``. Defaults to None, in which case default values for ``mentions``
            are used.
        normalize (bool, optional): Whether or not to normalize data. Defaults to False.
        profile_colors (dict[str, str], optional): Dictionary mapping profile labels
            to colors. Defaults to None, in which case we use a subset of ``COLOR_LIST``
            from ``utils`` module. Dictionary keys can be a subset of the profiles.
        bar_width (float, optional): Width of bars. Defaults to None which computes the bar width
            as 0.7 divided by the number of data sets. Must be in the interval :math:`(0,1]`.
        candidate_ordering (list[str], optional): Ordering of x-labels. Defaults to order retrieved
            from score dictionary.
        x_axis_name (str, optional): Name of x-axis. Defaults to None, which does not plot a name.
        y_axis_name (str, optional): Name of y-axis. Defaults to None, which does not plot a name.
        title (str, optional): Title for the figure. Defaults to None, which does not plot a title.
        show_profile_legend (bool, optional): Whether or not to plot the profile legend.
            Defaults to False. Is automatically shown if any threshold lines have the keyword
            "label" passed through ``threshold_kwds``.
        candidate_legend (dict[str, str], optional): Dictionary mapping candidates
            to relableing. Defaults to None. If provided, generates a second legend for data
            categories.
        relabel_candidates_with_int (bool, optional): Relabel the candidates with integer labels.
            Defaults to False. If ``candidate_legend`` is passed, those labels supercede.
        threshold_values (Union[list[float], float], optional): List of values to plot horizontal
            lines at. Can be provided as a list or a single float.
        threshold_kwds (Union[list[dict], dict], optional): List of plotting
            keywords for the horizontal lines. Can be a list or single dictionary. These will be
            passed to plt.axhline(). Common keywords include "linestyle", "linewidth", and "label".
            If "label" is passed, automatically plots the data set legend with the labels.
        legend_font_size (float, optional): The font size to use for the legend. Defaults to 10.0
            + the number of categories.
        legend_loc (str, optional): The location parameter to pass to ``Axes.legend(loc=)``.
            Defaults to "center left".
        legend_bbox_to_anchor (Tuple[float, float], otptional): The bounding box to anchor
            the legend to. Defaults to (1, 0.5).
        ax (Axes, optional): A matplotlib axes object to plot the figure on. Defaults to None, in
            which case the function creates and returns a new axes. The figure height is 6 inches
            and the figure width is 3 inches times the number of categories.

    Returns:
        Axes: A ``matplotlib`` axes with a bar plot of the given data.
    """
    stat_function = partial(mentions, **mentions_kwds) if mentions_kwds else mentions
    data_dict = _create_data_dict(profile_dict, stat_function)  # type: ignore[arg-type]

    if candidate_ordering is None:
        candidate_ordering = sorted(
            next(iter(data_dict.values())).keys(),
            reverse=True,
            key=lambda x: next(iter(data_dict.values()))[x],
        )

    if relabel_candidates_with_int and not candidate_legend:
        candidate_legend = {c: str(i + 1) for i, c in enumerate(candidate_ordering)}

    try:
        ax = multi_bar_plot(
            data_dict,
            normalize=normalize,
            data_set_colors=profile_colors,
            bar_width=bar_width,
            category_ordering=candidate_ordering,
            x_axis_name=x_axis_name,
            y_axis_name=y_axis_name,
            title=title,
            show_data_set_legend=show_profile_legend,
            categories_legend=candidate_legend,
            threshold_values=threshold_values,
            threshold_kwds=threshold_kwds,
            legend_font_size=legend_font_size,
            ax=ax,
        )

        return ax
    except Exception as e:
        if "Cannot plot more than" in str(e):
            raise ValueError(f"Cannot plot more than {len(COLOR_LIST)} profiles.")

        elif str(e) == "category_ordering must be the same length as sub-dictionaries.":
            raise ValueError(
                (
                    "candidate_ordering must be the same length as the dictionary "
                    f"produced by mentions: {len(next(iter(data_dict.values())))}."
                )
            )

        else:
            raise e


def multi_profile_fpv_plot(
    profile_dict: dict[str, RankProfile],
    fpv_kwds: Optional[dict[str, Any]] = None,
    normalize: bool = False,
    profile_colors: Optional[dict[str, str]] = None,
    bar_width: Optional[float] = None,
    candidate_ordering: Optional[list[str]] = None,
    x_axis_name: Optional[str] = None,
    y_axis_name: Optional[str] = None,
    title: Optional[str] = None,
    show_profile_legend: bool = False,
    candidate_legend: Optional[dict[str, str]] = None,
    relabel_candidates_with_int: bool = False,
    threshold_values: Optional[Union[list[float], float]] = None,
    threshold_kwds: Optional[Union[list[dict], dict]] = None,
    legend_font_size: Optional[float] = None,
    ax: Optional[Axes] = None,
) -> Axes:
    """
    Plot the first place votes for a collection of profiles. Wrapper for ``multi_profile_bar_plot``.

    Args:
        profile_dict (dict[str, RankProfile]): Keys are profile labels and values are
            profiles to plot statistics for.
        fpv_kwds (dict[str, Any], optional): Keyword arguments to pass to
            ``first_place_votes``. Defaults to None, in which case default values for
            ``first_place_votes`` are used.
        normalize (bool, optional): Whether or not to normalize data. Defaults to False.
        profile_colors (dict[str, str], optional): Dictionary mapping profile labels
            to colors. Defaults to None, in which case we use a subset of ``COLOR_LIST``
            from ``utils`` module. Dictionary keys can be a subset of the profiles.
        bar_width (float, optional): Width of bars. Defaults to None which computes the bar width
            as 0.7 divided by the number of data sets. Must be in the interval :math:`(0,1]`.
        candidate_ordering (list[str], optional): Ordering of x-labels. Defaults to order retrieved
            from score dictionary.
        x_axis_name (str, optional): Name of x-axis. Defaults to None, which does not plot a name.
        y_axis_name (str, optional): Name of y-axis. Defaults to None, which does not plot a name.
        title (str, optional): Title for the figure. Defaults to None, which does not plot a title.
        show_profile_legend (bool, optional): Whether or not to plot the profile legend.
            Defaults to False. Is automatically shown if any threshold lines have the keyword
            "label" passed through ``threshold_kwds``.
        candidate_legend (dict[str, str], optional): Dictionary mapping candidates
            to relableing. Defaults to None. If provided, generates a second legend for data
            categories.
        relabel_candidates_with_int (bool, optional): Relabel the candidates with integer labels.
            Defaults to False. If ``candidate_legend`` is passed, those labels supercede.
        threshold_values (Union[list[float], float], optional): List of values to plot horizontal
            lines at. Can be provided as a list or a single float.
        threshold_kwds (Union[list[dict], dict], optional): List of plotting
            keywords for the horizontal lines. Can be a list or single dictionary. These will be
            passed to plt.axhline(). Common keywords include "linestyle", "linewidth", and "label".
            If "label" is passed, automatically plots the data set legend with the labels.
        legend_font_size (float, optional): The font size to use for the legend. Defaults to 10.0
            + the number of categories.
        legend_loc (str, optional): The location parameter to pass to ``Axes.legend(loc=)``.
            Defaults to "center left".
        legend_bbox_to_anchor (Tuple[float, float], otptional): The bounding box to anchor
            the legend to. Defaults to (1, 0.5).
        ax (Axes, optional): A matplotlib axes object to plot the figure on. Defaults to None, in
            which case the function creates and returns a new axes. The figure height is 6 inches
            and the figure width is 3 inches times the number of categories.

    Returns:
        Axes: A ``matplotlib`` axes with a bar plot of the given data.
    """
    stat_function = (
        partial(first_place_votes, **fpv_kwds) if fpv_kwds else first_place_votes
    )
    data_dict = _create_data_dict(profile_dict, stat_function)  # type: ignore[arg-type]

    if candidate_ordering is None:
        candidate_ordering = sorted(
            next(iter(data_dict.values())).keys(),
            reverse=True,
            key=lambda x: next(iter(data_dict.values()))[x],
        )

    if relabel_candidates_with_int and not candidate_legend:
        candidate_legend = {c: str(i + 1) for i, c in enumerate(candidate_ordering)}

    try:
        ax = multi_bar_plot(
            data_dict,
            normalize=normalize,
            data_set_colors=profile_colors,
            bar_width=bar_width,
            category_ordering=candidate_ordering,
            x_axis_name=x_axis_name,
            y_axis_name=y_axis_name,
            title=title,
            show_data_set_legend=show_profile_legend,
            categories_legend=candidate_legend,
            threshold_values=threshold_values,
            threshold_kwds=threshold_kwds,
            legend_font_size=legend_font_size,
            ax=ax,
        )

        return ax
    except Exception as e:
        if "Cannot plot more than" in str(e):
            raise ValueError(f"Cannot plot more than {len(COLOR_LIST)} profiles.")

        elif str(e) == "category_ordering must be the same length as sub-dictionaries.":
            raise ValueError(
                (
                    "candidate_ordering must be the same length as the dictionary "
                    f"produced by first_place_votes: {len(next(iter(data_dict.values())))}."
                )
            )

        else:
            raise e


def multi_profile_ballot_lengths_plot(
    profile_dict: dict[str, RankProfile],
    ballot_lengths_kwds: Optional[dict[str, Any]] = None,
    normalize: bool = False,
    profile_colors: Optional[dict[str, str]] = None,
    bar_width: Optional[float] = None,
    lengths_ordering: Optional[list[str]] = None,
    x_axis_name: Optional[str] = None,
    y_axis_name: Optional[str] = None,
    title: Optional[str] = None,
    show_profile_legend: bool = False,
    lengths_legend: Optional[dict[str, str]] = None,
    threshold_values: Optional[Union[list[float], float]] = None,
    threshold_kwds: Optional[Union[list[dict], dict]] = None,
    legend_font_size: Optional[float] = None,
    ax: Optional[Axes] = None,
) -> Axes:
    """
    Plot the ballot lengths for a collection of profiles. Wrapper for ``multi_profile_bar_plot``.

    Args:
        profile_dict (dict[str, RankProfile]): Keys are profile labels and values are
            profiles to plot statistics for.
        ballot_lengths_kwds (dict[str, Any], optional): Keyword arguments to pass to
            ``ballot_lengths``. Defaults to None, in which case default values for
            ``ballot_lengths`` are used.
        normalize (bool, optional): Whether or not to normalize data. Defaults to False.
        profile_colors (dict[str, str], optional): Dictionary mapping profile labels
            to colors. Defaults to None, in which case we use a subset of ``COLOR_LIST``
            from ``utils`` module. Dictionary keys can be a subset of the profiles.
        bar_width (float, optional): Width of bars. Defaults to None which computes the bar width
            as 0.7 divided by the number of data sets. Must be in the interval :math:`(0,1]`.
        lengths_ordering (list[str], optional): Ordering of x-labels. Defaults to order retrieved
            from score dictionary.
        x_axis_name (str, optional): Name of x-axis. Defaults to None, which does not plot a name.
        y_axis_name (str, optional): Name of y-axis. Defaults to None, which does not plot a name.
        title (str, optional): Title for the figure. Defaults to None, which does not plot a title.
        show_profile_legend (bool, optional): Whether or not to plot the profile legend.
            Defaults to False. Is automatically shown if any threshold lines have the keyword
            "label" passed through ``threshold_kwds``.
        lengths_legend (dict[str, str], optional): Dictionary mapping lengths
            to relableing. Defaults to None. If provided, generates a second legend for data
            categories.
        threshold_values (Union[list[float], float], optional): List of values to plot horizontal
            lines at. Can be provided as a list or a single float.
        threshold_kwds (Union[list[dict], dict], optional): List of plotting
            keywords for the horizontal lines. Can be a list or single dictionary. These will be
            passed to plt.axhline(). Common keywords include "linestyle", "linewidth", and "label".
            If "label" is passed, automatically plots the data set legend with the labels.
        legend_font_size (float, optional): The font size to use for the legend. Defaults to 10.0
            + the number of categories.
        legend_loc (str, optional): The location parameter to pass to ``Axes.legend(loc=)``.
            Defaults to "center left".
        legend_bbox_to_anchor (Tuple[float, float], otptional): The bounding box to anchor
            the legend to. Defaults to (1, 0.5).
        ax (Axes, optional): A matplotlib axes object to plot the figure on. Defaults to None, in
            which case the function creates and returns a new axes. The figure height is 6 inches
            and the figure width is 3 inches times the number of categories.

    Returns:
        Axes: A ``matplotlib`` axes with a bar plot of the given data.
    """
    stat_function = (
        partial(ballot_lengths, **ballot_lengths_kwds)
        if ballot_lengths_kwds
        else ballot_lengths
    )
    data_dict = _create_data_dict(profile_dict, stat_function)  # type: ignore[arg-type]

    if lengths_ordering is None:
        lengths_ordering = sorted(
            next(iter(data_dict.values())).keys(),
            reverse=False,
            key=lambda x: x,
        )

    try:
        ax = multi_bar_plot(
            data_dict,
            normalize=normalize,
            data_set_colors=profile_colors,
            bar_width=bar_width,
            category_ordering=lengths_ordering,
            x_axis_name=x_axis_name,
            y_axis_name=y_axis_name,
            title=title,
            show_data_set_legend=show_profile_legend,
            categories_legend=lengths_legend,
            threshold_values=threshold_values,
            threshold_kwds=threshold_kwds,
            legend_font_size=legend_font_size,
            ax=ax,
        )

        return ax
    except Exception as e:
        if "Cannot plot more than" in str(e):
            raise ValueError(f"Cannot plot more than {len(COLOR_LIST)} profiles.")

        elif str(e) == "category_ordering must be the same length as sub-dictionaries.":
            raise ValueError(
                (
                    "lengths_ordering must be the same length as the dictionary "
                    f"produced by ballot_lengths: {len(next(iter(data_dict.values())))}."
                )
            )

        else:
            raise e



from votekit.ballot_generator.ballot_generator import BallotGenerator
from votekit.ballot_generator.utils import sample_cohesion_ballot_types
from votekit.ballot_generator.bloc_slate_generator import (
    name_bt_profile_generator,
    name_bt_profiles_by_bloc_generator,
    name_bt_profile_generator_using_mcmc,
    name_bt_profiles_by_bloc_generator_using_mcmc,
    name_pl_profile_generator,
    name_pl_profiles_by_bloc_generator,
    name_cumulative_profile_generator,
    name_cumulative_ballot_generator_by_bloc,
    cambridge_profile_generator,
    cambridge_profiles_by_bloc_generator,
    slate_pl_profile_generator,
    slate_pl_profiles_by_bloc_generator,
    slate_bt_profile_generator,
    slate_bt_profiles_by_bloc_generator,
    slate_bt_profile_generator_using_mcmc,
    slate_bt_profiles_by_bloc_generator_using_mcmc,
    convert_cohesion_map_to_cohesion_df,
    convert_preference_map_to_preference_df,
    convert_bloc_proportion_map_to_series,
    BlocSlateConfig,
)

from votekit.ballot_generator.std_generator import (
    ic_profile_generator,
    iac_profile_generator,
    onedim_spacial_profile_generator,
    spacial_profile_and_positions_generator,
    clustered_spacial_profile_and_positions_generator,
)

from votekit.ballot_generator.deprecated_classes import (
    name_BradleyTerry,
    name_PlackettLuce,
    short_name_PlackettLuce,
    name_Cumulative,
    CambridgeSampler,
    slate_PlackettLuce,
    slate_BradleyTerry,
    ImpartialCulture,
    ImpartialAnonymousCulture,
    Spatial,
    OneDimSpatial,
    ClusteredSpatial,
    AlternatingCrossover,
)


__all__ = [
    "BallotGenerator",
    "sample_cohesion_ballot_types",
    "name_bt_profile_generator",
    "name_bt_profiles_by_bloc_generator",
    "name_bt_profile_generator_using_mcmc",
    "name_bt_profiles_by_bloc_generator_using_mcmc",
    "name_pl_profile_generator",
    "name_pl_profiles_by_bloc_generator",
    "name_cumulative_profile_generator",
    "name_cumulative_ballot_generator_by_bloc",
    "cambridge_profile_generator",
    "cambridge_profiles_by_bloc_generator",
    "slate_pl_profile_generator",
    "slate_pl_profiles_by_bloc_generator",
    "slate_bt_profile_generator",
    "slate_bt_profiles_by_bloc_generator",
    "slate_bt_profile_generator_using_mcmc",
    "slate_bt_profiles_by_bloc_generator_using_mcmc",
    "ic_profile_generator",
    "iac_profile_generator",
    "onedim_spacial_profile_generator",
    "spacial_profile_and_positions_generator",
    "clustered_spacial_profile_and_positions_generator",
    "convert_cohesion_map_to_cohesion_df",
    "convert_preference_map_to_preference_df",
    "convert_bloc_proportion_map_to_series",
    "BlocSlateConfig",
    "name_BradleyTerry",
    "name_PlackettLuce",
    "short_name_PlackettLuce",
    "name_Cumulative",
    "CambridgeSampler",
    "slate_PlackettLuce",
    "slate_BradleyTerry",
    "ImpartialCulture",
    "ImpartialAnonymousCulture",
    "Spatial",
    "OneDimSpatial",
    "ClusteredSpatial",
    "AlternatingCrossover",
]



from matplotlib import pyplot as plt  # type: ignore
from votekit.utils import COLOR_LIST
from matplotlib.axes import Axes
from typing import Optional, Union, Tuple, Literal, Any
import matplotlib.patches as mpatches
import warnings
from matplotlib.legend import Legend
from matplotlib.lines import Line2D

DEFAULT_LINE_KWDS = {"linestyle": "-", "linewidth": 2, "color": "grey", "alpha": 0.5}


def add_null_keys(data: dict[str, dict[str, float]]) -> dict[str, dict[str, float]]:
    """
    Prepares dictionary of dictionaries to be passed to ``multi_bar_plot()``. If a key is missing
    from a dictionary, this function adds the key with value 0.

    Args:
        data (dict[str, dict[str, float]]): Categorical data to be cleaned. The value of each dict
            should be the frequency of the key which is the category name.

    Returns:
        dict[str, dict[str, float]]: Cleaned data.
    """

    x_labels = []
    for data_dict in data.values():
        for x_label in data_dict.keys():
            if x_label not in x_labels:
                x_labels.append(x_label)

    for x_label in x_labels:
        for data_dict in data.values():
            if x_label not in data_dict:
                data_dict[x_label] = 0

    return data


def _set_default_bar_plot_args(
    *,
    data: dict[str, dict[str, float]],
    data_set_colors: Optional[dict],
    bar_width: Optional[float],
    category_ordering: Optional[list[str]],
    legend_font_size: Optional[float],
    threshold_values: Optional[Union[list[float], float]],
    threshold_kwds: Optional[Union[list[dict], dict]],
    ax: Optional[Axes],
) -> dict[str, Any]:
    """
    Handles setting default arguments.

    Args:
        data (dict[str, dict[str, float]]): Categorical data to be plotted. Top level keys are
            data set labels. Inner keys are categories, and inner values are the height of the bars.
        data_set_colors (dict, optional): Dictionary mapping data set labels or data set indices to
            colors. Defaults to None, in which case we use a subset of ``COLOR_LIST`` from ``utils``
            module. Dictionary keys can be a subset of the data sets.
        bar_width (float, optional): Width of bars, defaults to None which computes the bar width
            as 0.7 divided by the number of data sets.
        category_ordering (list[str], optional): Ordering of x labels. Defaults to order retrieved
            from first data dictionary.
        threshold_values (Union[list[float], float], optional): List of values to plot horizontal
            lines at. Can be provided as a list of values or a single float.
        threshold_kwds (Union[list[dict], dict], optional): List of plotting
            keywords for the horizontal lines. Can be a list of dictionaries or single dictionary.
        ax (Axes, optional): A matplotlib axes object to plot the figure on. Defaults to None, in
            which case the function creates and returns a new axes.

    Returns:
        dict[str, Any]: data_set_to_color, bar_width, category_ordering, font_size,
        legend_font_size, threshold_values, threshold_kwds, ax
    """

    data_set_to_color = {
        d: COLOR_LIST[i % len(COLOR_LIST)] for i, d in enumerate(data.keys())
    }

    if data_set_colors:
        for data_set, color in data_set_colors.items():
            data_set_to_color[data_set] = color

    if bar_width is None:
        bar_width = 0.7 / len(data)
    else:
        bar_width *= 1 / len(data)

    if category_ordering is None:
        category_ordering = list(next(iter(data.values())).keys())

    FONT_SIZE = len(category_ordering) + 10.0

    if legend_font_size is None:
        legend_font_size = FONT_SIZE

    if threshold_values is not None and not isinstance(threshold_values, list):
        threshold_values = [threshold_values]

    if threshold_kwds is not None and not isinstance(threshold_kwds, list):
        threshold_kwds = [threshold_kwds]

    if isinstance(threshold_values, list):
        if not isinstance(threshold_kwds, list):
            threshold_kwds = [dict() for _ in range(len(threshold_values))]
        for kwd_dict in threshold_kwds:
            for k, v in DEFAULT_LINE_KWDS.items():
                if k not in kwd_dict:
                    kwd_dict[k] = v

    if ax is None:
        _, ax = plt.subplots(figsize=(3 * len(category_ordering), 6))

    return {
        "data_set_to_color": data_set_to_color,
        "bar_width": bar_width,
        "category_ordering": category_ordering,
        "font_size": FONT_SIZE,
        "legend_font_size": legend_font_size,
        "threshold_values": threshold_values,
        "threshold_kwds": threshold_kwds,
        "ax": ax,
    }


def _validate_bar_plot_args(
    *,
    data: dict[str, dict[str, float]],
    category_ordering: list[str],
    categories_legend: Optional[dict[str, str]],
    bar_width: float,
    threshold_values: Optional[list[float]],
    threshold_kwds: Optional[list[dict[str, str]]],
) -> float:
    """
    Validates bar plot arguments.

    Args:
        data (dict[str, dict[str, float]]): Categorical data to be plotted. Top level keys are
            data set labels. Inner keys are categories, and inner values are the height of the bars.
        category_ordering (list[str]): Ordering of x labels. Must match data keys.
        categories_legend (dict[str, str], optional): Dictionary mapping data categories
            to relabeling. If provided, generates a second legend for data
            categories and relabels the x-axis accordingly. Can be a subset of the data keys.
        bar_width (float): Width of bars. Must be between 0 and 1/len(data).
        threshold_values (list[float], optional): List of values to plot horizontal
            lines at.
        threshold_kwds (list[dict[str, str]], optional): List of plotting
            keywords for the horizontal lines.

    Raises:
        ValueError: All bar heights must be positive.
        ValueError: All sub dictionaries in data must have the same keys.
        ValueError: Cannot plot more than len(COLOR_LIST) data sets.
        ValueError: category_ordering must be the same length as the keys of sub-dictionaries.
        ValueError: category_ordering must have no extraneous labels.
        ValueError: category_ordering must have no missing labels.
        ValueError: All keys in category_legend must be a subset of data keys.
        ValueError: Bar width must be positive.
        UserWarning: Bar width must be less than 1.
        ValueError: threshold_values and kwds must have the same length.
        ValueError: If lw or ls is passed as a threshold keyword.

    Returns:
        float: Bar width.
    """
    if any(v < 0 for data_dict in data.values() for v in data_dict.values()):
        raise ValueError("All bar heights must be non-negative.")

    if any(
        set(next(iter(data.values())).keys()) != set(data_dict.keys())
        for data_dict in data.values()
    ):
        raise ValueError("All data dictionaries must have the same categorical keys.")

    if len(data) > len(COLOR_LIST):
        raise ValueError(f"Cannot plot more than {len(COLOR_LIST)} data sets.")

    true_categories = next(iter(data.values())).keys()

    if len(category_ordering) != len(true_categories):
        raise ValueError(
            "category_ordering must be the same length as sub-dictionaries."
        )

    if set(category_ordering).difference(true_categories) != set():
        raise ValueError(
            (
                "category_ordering has extraneous labels: "
                f"{set(category_ordering).difference(true_categories)}"
            )
        )

    if set(true_categories).difference(category_ordering) != set():
        raise ValueError(
            (
                "category_ordering has missing labels: "
                f"{set(true_categories).difference(category_ordering)}"
            )
        )

    if categories_legend:
        mislabeled_keys = set(categories_legend.keys()).difference(true_categories)

        if mislabeled_keys != set():
            raise ValueError(
                f"{mislabeled_keys} in categories_legend must "
                f"be a subset of {set(true_categories)}"
            )
    if bar_width <= 0:
        raise ValueError("Bar width must be positive.")

    elif bar_width > 1 / len(data):
        warnings.warn(
            "Bar width must be less than 1. Bar width has now been set to 1.",
            category=UserWarning,
        )
        bar_width = 1 / len(data)

    if threshold_values and threshold_kwds:
        if len(threshold_kwds) != len(threshold_values):
            raise ValueError(
                "threshold_values must have the same length as threshold_kwds."
            )
        if any("ls" in d.keys() for d in threshold_kwds):
            raise ValueError("Must use linestyle, not ls.")
        if any("lw" in d.keys() for d in threshold_kwds):
            raise ValueError("Must use linewidth, not lw.")

    return bar_width


def _normalize_data_dict(data_dict: dict[str, float]) -> dict[str, float]:
    """
    Normalizes data so number of total observations is 1. Raises a ValueError if the total
    mass is 0.

    Args:
        data (dict[str, float]): Single data dictionary whose keys are categories and values
            are bar heights.

    Returns:
        dict[str, float]: Normalized data.

    Raise:
        ValueError: If the sum of the data is 0.
    """

    total_obs = sum(data_dict.values())

    if total_obs == 0:
        raise ValueError("Total mass of observations must be non-zero.")

    return {k: v / total_obs for k, v in data_dict.items()}


def _prepare_data_bar_plot(
    *,
    normalize: bool,
    data: dict[str, dict[str, float]],
    category_ordering: list[str],
) -> tuple[list[list[float]], float]:
    """
    Formats data and normalizes if required.

    Args:
        normalize (bool): Whether or not to normalize data.
        data (dict[str, dict[str, float]]): Categorical data to be plotted. Top level keys are
            data set labels. Inner keys are categories, and inner values are the height of the bars.
        category_ordering (list[str]): Ordering of x labels.

    Returns:
        list[list[float]]: Height of bars, one list for each data set.

    """
    plot_data = data
    if normalize:
        plot_data = {
            label: _normalize_data_dict(data_dict) for label, data_dict in data.items()
        }

    y_data = [
        [data_dict[x_label] for x_label in category_ordering]
        for data_dict in plot_data.values()
    ]

    all_data_values: list[float] = []
    # NOTE: data is dict[str, dict[str, float]]
    for dct in data.values():
        if isinstance(dct, dict):
            all_data_values.extend(dct.values())

    flat_y_data = [item for sublist in y_data for item in sublist]

    data_ratio = float(
        1.0 if not normalize else max(flat_y_data) / max(all_data_values)
    )

    assert isinstance(
        data_ratio, float
    ), "Something went wrong in computation of data_ratio."

    return y_data, data_ratio


def _plot_datasets_on_bar_plot(
    *,
    ax: Axes,
    category_ordering: list[str],
    y_data: list[list[float]],
    data_set_labels: list[str],
    categories_legend: Optional[dict[str, str]],
    bar_width: float,
    data_set_to_color: dict,
    font_size: float,
) -> Axes:
    """

    Args:
        category_ordering (list[str]): Ordering of x labels.
        y_data (list[list[floats]]): List of lists where each sublist is the bar heights for a
            data set.
        data_set_labels (list[str]): List of labels for data sets.
        categories_legend (dict[str, str], optional): Dictionary mapping data categories
            to relabeling. If provided, generates a second legend for data
            categories and relabels the x-axis accordingly. Can be a subset of the data keys.
        bar_width (float): Width of bars.
        data_set_to_color (dict): Dictionary mapping data set labels to colors.
        font_size (float): Font size for figure.

    Returns:
        Axes: Matplotlib axes containing bar plot.
    """
    if categories_legend:
        tick_labels = [categories_legend.get(c, c) for c in category_ordering]
    else:
        tick_labels = category_ordering

    for i, data_set in enumerate(y_data):
        bar_shift = bar_width * (i - len(y_data) / 2)
        bar_centers = [x + bar_shift for x in range(len(category_ordering))]

        align: Literal["center", "edge"] = "edge" if len(y_data) % 2 == 0 else "center"

        if i == len(y_data) // 2:
            ax.bar(
                bar_centers,
                data_set,
                tick_label=tick_labels,
                label=data_set_labels[i],
                align=align,
                width=bar_width,
                color=data_set_to_color[data_set_labels[i]],
            )
        else:
            ax.bar(
                bar_centers,
                data_set,
                label=data_set_labels[i],
                align=align,
                width=bar_width,
                color=data_set_to_color[data_set_labels[i]],
            )

    ax.tick_params(axis="x", labelsize=font_size)
    ax.tick_params(axis="y", labelsize=font_size)

    if len(y_data) % 2 == 0:
        lower_bar_edge = -bar_width / 2 * len(y_data)
        upper_bar_edge = len(category_ordering) - 1 + 0.5 * bar_width * len(y_data)

    else:
        lower_bar_edge = -bar_width / 2 * (1 + len(y_data))
        upper_bar_edge = len(category_ordering) - 1 + bar_width / 2 * (len(y_data) - 1)

    gap_size = 0.25 / len(y_data)
    lower = lower_bar_edge - gap_size
    upper = upper_bar_edge + gap_size
    ax.set_xbound(lower, upper)

    return ax


def _label_bar_plot(
    *,
    ax: Axes,
    x_axis_name: Optional[str],
    y_axis_name: Optional[str],
    title: Optional[str],
    font_size: float,
) -> Axes:
    """
    Add x,y axes labels and title to bar plot.

    Args:
        ax (Axes): Matplotlib axes containing barplot.
        x_axis_name (str, optional): Name of x-axis.
        y_axis_name (str, optional): Name of y-axis.
        title (str, optional): Title for the figure.
        font_size (float): Font size for figure.

    Returns:
        Axes: Labeled bar plot.

    """
    if x_axis_name:
        ax.set_xlabel(x_axis_name, fontsize=font_size)
    if y_axis_name:
        ax.set_ylabel(y_axis_name, fontsize=font_size)
    if title:
        ax.set_title(title, fontsize=font_size)

    return ax


def _add_horizontal_lines_bar_plot(
    *,
    threshold_values: list[float],
    threshold_kwds: list[dict],
    ax: Axes,
    data_ratio: float = 1.0,
):
    """
    Add horizontal lines to the bar plot.

    Args:
        threshold_values (list[float]): Display a horizontal line at given y-values.
        threshold_kwds (list[dict]): Matplotlib keywords for plotting lines.
        ax (Axes): Axes to plot on.

    Returns:
        Axes: Axes with horizontal lines added.

    """

    for i, y in enumerate(threshold_values):
        ax.axhline(y * data_ratio, **threshold_kwds[i])

    return ax


def _add_data_sets_legend_bar_plot(
    *,
    ax: Axes,
    data_set_labels: list[str],
    data_set_to_color: dict,
    categories_legend: Optional[dict[str, str]],
    threshold_kwds: Optional[list[dict]],
    legend_font_size: float,
    legend_loc: str,
    legend_bbox_to_anchor: Tuple[float, float],
) -> Tuple[Axes, Optional[Legend]]:
    """
    Add legend to bar plot for data sets and any horizontal lines.

    Args:
        ax (Axes): Matplotlib axes containing barplot.
        data_set_labels (list[str]): Labels for each data set.
        data_set_to_color (dict): Dictionary mapping data set labels to colors.
        categories_legend (dict[str, str], optional): Dictionary mapping x-axis
            categories to description in legend. Defaults to None, in which case legend is just
            x-axis labels.
        threshold_kwds (list[dict], optional): List of plotting
            keywords for the horizontal lines.
        legend_font_size (float): The font size to use for the legend.
        legend_loc(str): The location parameter to pass to ``Axes.legend(loc=)``.
        legend_bbox_to_anchor (Tuple[float, float]): The bounding box to anchor the legend to.

    Returns:
        Tuple[Axes, Legend]: Matplotlib axes containing bar plot with legend for data sets.
    """
    proxy_artists: list[Union[mpatches.Patch, Line2D]] = []

    for label in data_set_labels:
        patch = mpatches.Patch(color=data_set_to_color[label], label=label)
        proxy_artists.append(patch)

    if threshold_kwds:
        for kwd_dict in threshold_kwds:
            line = Line2D([0], [0], **kwd_dict)
            proxy_artists.append(line)

    leg = ax.get_legend()
    if proxy_artists:
        leg = ax.legend(
            handles=proxy_artists,
            loc=legend_loc,
            bbox_to_anchor=legend_bbox_to_anchor,
            fontsize=legend_font_size,
            ncol=len(proxy_artists) // 15 + 1,
            frameon=True,
            fancybox=True,
        )

    if categories_legend and leg is not None:
        ax.add_artist(leg)

    return ax, leg


def _add_categories_legend_bar_plot(
    *,
    ax: Axes,
    categories_legend: dict[str, str],
    legend_font_size: float,
    legend_loc: str,
    legend_bbox_to_anchor: Tuple[float, float],
    data_set_legend: Optional[Legend],
) -> Axes:
    """
    Add legend to bar plot.

    Args:
        ax (Axes): Matplotlib axes containing barplot.
        categories_legend (dict[str, str]): Dictionary mapping x-axis
            categories to description in legend.
        legend_font_size (float): The font size to use for the legend.
        legend_loc(str): The location parameter to pass to ``Axes.legend(loc=)``.
        legend_bbox_to_anchor (Tuple[float, float]): The bounding box to anchor the legend to.
        data_set_legend (Legend, optional): The data set legend. Defaults to None.


    Returns:
        Axes: Matplotlib axes containing bar plot with legend.
    """
    proxy_artists = []

    for category, relabel in categories_legend.items():
        patch = mpatches.Patch(color="white", label=f"{relabel}: {category}")
        proxy_artists.append(patch)

    if proxy_artists:
        if data_set_legend:
            bbox = data_set_legend.get_window_extent()
            fig = plt.gcf()
            # fig.canvas.draw()

            width = bbox.transformed(fig.transFigure.inverted()).width
            legend_bbox_to_anchor = (
                legend_bbox_to_anchor[0] + 1.4 * width,
                legend_bbox_to_anchor[1],
            )

        ax.legend(
            handles=proxy_artists,
            loc=legend_loc,
            bbox_to_anchor=legend_bbox_to_anchor,
            fontsize=legend_font_size,
            ncol=len(proxy_artists) // 15 + 1,
            handlelength=0,
            handletextpad=0,
            frameon=True,
            fancybox=True,
        )

    return ax


def multi_bar_plot(
    data: dict[str, dict[str, float]],
    *,
    normalize: bool = False,
    data_set_colors: Optional[dict[str, str]] = None,
    bar_width: Optional[float] = None,
    category_ordering: Optional[list[str]] = None,
    x_axis_name: Optional[str] = None,
    y_axis_name: Optional[str] = None,
    title: Optional[str] = None,
    show_data_set_legend: bool = False,
    categories_legend: Optional[dict[str, str]] = None,
    threshold_values: Optional[Union[list[float], float]] = None,
    threshold_kwds: Optional[Union[list[dict], dict]] = None,
    legend_font_size: Optional[float] = None,
    ax: Optional[Axes] = None,
) -> Axes:
    """
    Plots bar plot of categorical data.

    Args:
        data (dict[str, dict[str, float]]): Categorical data to be plotted. Top level keys are
            data set labels. Inner keys are categories, and inner values are the height of the bars.
        normalize (bool, optional): Whether or not to normalize data. Defaults to False.
        data_set_colors (dict[str, str], optional): Dictionary mapping data set labels
            to colors. Defaults to None, in which case we use a subset of ``COLOR_LIST``
            from ``utils`` module. Dictionary keys can be a subset of the data sets.
        bar_width (float, optional): Width of bars. Defaults to None which computes the bar width
            as 0.7 divided by the number of data sets. Must be in the interval :math:`(0,1]`.
        category_ordering (list[str], optional): Ordering of x-labels. Defaults to order retrieved
            from data dictionary.
        x_axis_name (str, optional): Name of x-axis. Defaults to None, which does not plot a name.
        y_axis_name (str, optional): Name of y-axis. Defaults to None, which does not plot a name.
        title (str, optional): Title for the figure. Defaults to None, which does not plot a title.
        show_data_set_legend (bool, optional): Whether or not to plot the data set legend.
            Defaults to False. Is automatically shown if any threshold lines have the keyword
            "label" passed through ``threshold_kwds``.
        categories_legend (dict[str, str], optional): Dictionary mapping data categories
            to relabeling. Defaults to None. If provided, generates a second legend for data
            categories and relabels the x-axis accordingly. Can be a subset of the data keys.
        threshold_values (Union[list[float], float], optional): List of values to plot horizontal
            lines at. Can be provided as a list or a single float.
        threshold_kwds (Union[list[dict], dict], optional): List of plotting
            keywords for the horizontal lines. Can be a list or single dictionary. These will be
            passed to plt.axhline(). Common keywords include "linestyle", "linewidth", and "label".
            If "label" is passed, automatically plots the data set legend with the labels.
        legend_font_size (float, optional): The font size to use for the legend. Defaults to 10.0
            + the number of categories.
        legend_loc (str, optional): The location parameter to pass to ``Axes.legend(loc=)``.
            Defaults to "center left".
        legend_bbox_to_anchor (Tuple[float, float], otptional): The bounding box to anchor
            the legend to. Defaults to (1, 0.5).
        ax (Axes, optional): A matplotlib axes object to plot the figure on. Defaults to None, in
            which case the function creates and returns a new axes. The figure height is 6 inches
            and the figure width is 3 inches times the number of categories.

    Returns:
        Axes: A ``matplotlib`` axes with a bar plot of the given data.
    """

    barplot_kwds = _set_default_bar_plot_args(
        data=data,
        data_set_colors=data_set_colors,
        bar_width=bar_width,
        category_ordering=category_ordering,
        legend_font_size=legend_font_size,
        threshold_values=threshold_values,
        threshold_kwds=threshold_kwds,
        ax=ax,
    )

    barplot_kwds["bar_width"] = _validate_bar_plot_args(
        data=data,
        category_ordering=barplot_kwds["category_ordering"],
        categories_legend=categories_legend,
        bar_width=barplot_kwds["bar_width"],
        threshold_values=barplot_kwds["threshold_values"],
        threshold_kwds=barplot_kwds["threshold_kwds"],
    )

    y_data, data_ratio = _prepare_data_bar_plot(
        normalize=normalize,
        data=data,
        category_ordering=barplot_kwds["category_ordering"],
    )

    ax = _plot_datasets_on_bar_plot(
        ax=barplot_kwds["ax"],
        category_ordering=barplot_kwds["category_ordering"],
        y_data=y_data,
        data_set_labels=list(data.keys()),
        categories_legend=categories_legend,
        bar_width=barplot_kwds["bar_width"],
        data_set_to_color=barplot_kwds["data_set_to_color"],
        font_size=barplot_kwds["font_size"],
    )

    ax = _label_bar_plot(
        ax=ax,
        x_axis_name=x_axis_name,
        y_axis_name=y_axis_name,
        title=title,
        font_size=barplot_kwds["font_size"],
    )

    if barplot_kwds["threshold_values"] and barplot_kwds["threshold_kwds"]:
        ax = _add_horizontal_lines_bar_plot(
            threshold_values=barplot_kwds["threshold_values"],
            threshold_kwds=barplot_kwds["threshold_kwds"],
            ax=ax,
            data_ratio=data_ratio,
        )

        for kwd_dict in barplot_kwds["threshold_kwds"]:
            if "label" in kwd_dict:
                show_data_set_legend = True
                break

    data_set_legend = None

    if show_data_set_legend:
        ax, data_set_legend = _add_data_sets_legend_bar_plot(
            ax=ax,  # type: ignore[arg-type]
            data_set_labels=list(data.keys()),
            data_set_to_color=barplot_kwds["data_set_to_color"],
            categories_legend=categories_legend,
            threshold_kwds=barplot_kwds["threshold_kwds"],
            legend_font_size=barplot_kwds["legend_font_size"],
            legend_loc="center left",
            legend_bbox_to_anchor=(1, 1 / 2),
        )

    if categories_legend:
        ax = _add_categories_legend_bar_plot(
            ax=ax,  # type: ignore[arg-type]
            categories_legend=categories_legend,
            legend_font_size=barplot_kwds["legend_font_size"],
            legend_loc="center left",
            legend_bbox_to_anchor=(1, 1 / 2),
            data_set_legend=data_set_legend,
        )

    return ax  # type: ignore[return-value]


def bar_plot(
    data: dict[str, float],
    *,
    data_set_label: str = "Data set",
    normalize: bool = False,
    data_set_color: str = COLOR_LIST[0],
    bar_width: Optional[float] = None,
    category_ordering: Optional[list[str]] = None,
    x_axis_name: Optional[str] = None,
    y_axis_name: Optional[str] = None,
    title: Optional[str] = None,
    show_data_set_legend: bool = False,
    categories_legend: Optional[dict[str, str]] = None,
    threshold_values: Optional[Union[list[float], float]] = None,
    threshold_kwds: Optional[Union[list[dict], dict]] = None,
    legend_font_size: Optional[float] = None,
    ax: Optional[Axes] = None,
) -> Axes:
    """
    Plots bar plot of a single categorical data set. Wrapper for ``multi_bar_plot``.

    Args:
        data (dict[str, float]): Categorical data set to be plotted. Keys are categories, and
            values are the height of the bars.
        data_set_label(str, optional): Label for data set. Defaults to "Data set".
        normalize (bool, optional): Whether or not to normalize data. Defaults to False.
        data_set_color (str, optional): Color of data set.
            Defaults to the first color from ``COLOR_LIST`` from ``utils`` module.
        bar_width (float, optional): Width of bars. Defaults to None which computes the bar width
            as 0.7 divided by the number of data sets. Must be in the interval :math:`(0,1]`.
        category_ordering (list[str], optional): Ordering of x-labels. Defaults to order retrieved
            from data dictionary.
        x_axis_name (str, optional): Name of x-axis. Defaults to None, which does not plot a name.
        y_axis_name (str, optional): Name of y-axis. Defaults to None, which does not plot a name.
        title (str, optional): Title for the figure. Defaults to None, which does not plot a title.
        show_data_set_legend (bool, optional): Whether or not to plot the data set legend.
            Defaults to False. Is automatically shown if any threshold lines have the keyword
            "label" passed through ``threshold_kwds``.
        categories_legend (dict[str, str], optional): Dictionary mapping data categories
            to description. Defaults to None. If provided, generates a second legend for data
            categories.
        threshold_values (Union[list[float], float], optional): List of values to plot horizontal
            lines at. Can be provided as a list or a single float.
        threshold_kwds (Union[list[dict], dict], optional): List of plotting
            keywords for the horizontal lines. Can be a list or single dictionary. These will be
            passed to plt.axhline(). Common keywords include "linestyle", "linewidth", and "label".
            If "label" is passed, automatically plots the data set legend with the labels.
        legend_font_size (float, optional): The font size to use for the legend. Defaults to 10.0
            + the number of categories.
        legend_loc (str, optional): The location parameter to pass to ``Axes.legend(loc=)``.
            Defaults to "center left".
        legend_bbox_to_anchor (Tuple[float, float], otptional): The bounding box to anchor
            the legend to. Defaults to (1, 0.5).
        ax (Axes, optional): A matplotlib axes object to plot the figure on. Defaults to None, in
            which case the function creates and returns a new axes. The figure height is 6 inches
            and the figure width is 3 inches times the number of categories.

    Returns:
        Axes: A ``matplotlib`` axes with a bar plot of the given data.
    """

    return multi_bar_plot(
        {data_set_label: data},
        normalize=normalize,
        data_set_colors={data_set_label: data_set_color},
        bar_width=bar_width,
        category_ordering=category_ordering,
        x_axis_name=x_axis_name,
        y_axis_name=y_axis_name,
        title=title,
        show_data_set_legend=show_data_set_legend,
        categories_legend=categories_legend,
        threshold_values=threshold_values,
        threshold_kwds=threshold_kwds,
        legend_font_size=legend_font_size,
        ax=ax,
    )



from votekit.elections.election_types.ranking.abstract_ranking import RankingElection
from votekit.pref_profile import RankProfile
from votekit.elections.election_state import ElectionState
from votekit.utils import first_place_votes
from votekit.cleaning import remove_and_condense_rank_profile
from votekit.elections.election_types.ranking import Plurality
from typing import Optional, Literal
from functools import partial


class TopTwo(RankingElection):
    """
    Eliminates all but the top two plurality vote-getters, and then conducts a runoff between them,
    reallocating other ballots.

    Args:
        profile (RankProfile): Profile to conduct election on.
        tiebreak (str, optional): Tiebreak method to use. Options are None, 'random', and 'borda'.
            Defaults to None, in which case a tie raises a ValueError.
        fpv_tie_convention (Literal["high", "average", "low"], optional): How to award points
            for tied first place votes. Defaults to "average", where if n candidates are tied for
            first, each receives 1/n points. "high" would award them each one point, and "low" 0.
            Only used by ``score_function`` parameter.

    """

    def __init__(
        self,
        profile: RankProfile,
        tiebreak: Optional[str] = None,
        fpv_tie_convention: Literal["high", "low", "average"] = "average",
    ):
        self.tiebreak = tiebreak
        super().__init__(
            profile,
            score_function=partial(
                first_place_votes, tie_convention=fpv_tie_convention
            ),
            sort_high_low=True,
        )

    def _is_finished(self):
        # two round election
        if len(self.election_states) == 3:
            return True
        return False

    def _run_step(
        self, profile: RankProfile, prev_state: ElectionState, store_states=False
    ) -> RankProfile:
        """
        Run one step of an election from the given profile and previous state.
        Choose top two candidates by first-place votes, then remove all other candidates
        from ballots and re-run to elect single winner by first-place votes.

        Args:
            profile (RankProfile): Profile of ballots.
            prev_state (ElectionState): The previous ElectionState.
            store_states (bool, optional): True if `self.election_states` should be updated with the
                ElectionState generated by this round. This should only be True when used by
                `self._run_election()`. Defaults to False.

        Returns:
            RankProfile: The profile of ballots after the round is completed.
        """
        if prev_state.round_number == 0:
            plurality = Plurality(profile, 2, self.tiebreak)

            remaining = plurality.get_elected()
            eliminated = plurality.get_remaining()
            tiebreaks = plurality.election_states[-1].tiebreaks
            new_profile: RankProfile = remove_and_condense_rank_profile(
                [c for s in eliminated for c in s], profile
            )

            if self.score_function is None:
                raise ValueError(
                    "score_function must be set for TopTwo elections when recording states"
                )

            scores = self.score_function(new_profile)

            if store_states:
                new_state = ElectionState(
                    round_number=prev_state.round_number + 1,
                    remaining=remaining,
                    eliminated=eliminated,
                    scores=scores,
                    tiebreaks=tiebreaks,
                )

                self.election_states.append(new_state)

        else:
            plurality = Plurality(profile, 1, self.tiebreak)
            new_profile = plurality.get_profile()

            if store_states:
                # first state was already stored by round 1 plurality
                # need to update round numbers
                plurality.election_states[1].round_number = 2
                self.election_states.append(plurality.election_states[1])

        return new_profile



import itertools as it
import random
import numpy as np
from typing import Union, Mapping
import psutil


def system_memory() -> dict[str, float]:
    """
    Returns a dictionary with system memory information in GiB via psutil.

    Returns:
        dict[str, float]: A dictionary with keys 'total_gib', 'available_gib',
            'used_gib', and 'percent' representing the total, available, and used
            memory in GiB and the percentage of used memory.
    """
    vm = psutil.virtual_memory()
    return {
        "total_gib": vm.total / 2**30,
        "available_gib": vm.available / 2**30,
        "used_gib": vm.used / 2**30,
        "percent": vm.percent,
    }


# TODO: Fix this up to be more readable. Also make sure to mention keys of
# cohesion_parameters_for_bloc are slates now.
def sample_cohesion_ballot_types(
    slate_to_non_zero_candidates: dict[str, list[str]],
    num_ballots: int,
    cohesion_parameters_for_bloc: Mapping[str, Union[float, int]],
) -> list[list[str]]:
    """
    Returns a list of ballots; each ballot is a list of slate names (strings)
    in the order they appear on that ballot.

    Args:
        slate_to_non_zero_candidates (dict[str, list[str]]):
        num_ballots (int):
        cohesion_parameters_for_bloc (Mapping[str, Union[float, int]]):

    Returns:
        list[list[str]]: A list of lists, where each list contains the bloc names in the order
            they appear on the ballot.
    """
    candidates = list(it.chain.from_iterable(slate_to_non_zero_candidates.values()))

    ballots: list[list[str]] = [[] for _ in range(num_ballots)]

    coin_flips = list(np.random.uniform(size=len(candidates) * num_ballots))

    def which_bin(dist_bins: list[float], flip: float) -> int:
        for i, left in enumerate(dist_bins[:-1]):
            if left < flip <= dist_bins[i + 1]:
                return i
        return len(dist_bins) - 2

    blocs_og, values_og = [list(x) for x in zip(*cohesion_parameters_for_bloc.items())]

    for j in range(num_ballots):
        blocs = blocs_og.copy()
        values = values_og.copy()

        distribution_bins: list[float] = [0.0] + [
            sum(values[: i + 1]) for i in range(len(blocs))
        ]
        ballot_type: list[str] = [""] * len(candidates)

        for i, flip in enumerate(
            coin_flips[j * len(candidates) : (j + 1) * len(candidates)]
        ):
            bloc_index = which_bin(distribution_bins, float(flip))
            bloc_type = blocs[bloc_index]
            ballot_type[i] = bloc_type

            if ballot_type.count(bloc_type) == len(
                slate_to_non_zero_candidates[bloc_type]
            ):
                del blocs[bloc_index]
                del values[bloc_index]
                total_value_sum = sum(values)

                if total_value_sum == 0 and len(values) > 0:
                    # remaining blocs have zero cohesion → fill with random permutation
                    remaining_blocs = [
                        b
                        for b in blocs
                        for _ in range(len(slate_to_non_zero_candidates[b]))
                    ]
                    random.shuffle(remaining_blocs)
                    ballot_type[i + 1 :] = remaining_blocs
                    break

                # renormalize and recompute bins
                values = [v / total_value_sum for v in values]
                distribution_bins = [0.0] + [
                    sum(values[: k + 1]) for k in range(len(blocs))
                ]

        ballots[j] = ballot_type

    return ballots



from .rank_profiles_cleaning import (
    clean_rank_profile,
    remove_repeat_cands_rank_profile,
    remove_cand_rank_profile,
    condense_rank_profile,
    remove_and_condense_rank_profile,
)

from .rank_ballots_cleaning import (
    remove_cand_rank_ballot,
    condense_rank_ballot,
    remove_repeat_cands_rank_ballot,
)

from .score_ballots_cleaning import remove_cand_score_ballot
from .score_profiles_cleaning import clean_score_profile, remove_cand_score_profile

__all__ = [
    "clean_rank_profile",
    "remove_repeat_cands_rank_profile",
    "remove_cand_rank_profile",
    "condense_rank_profile",
    "remove_and_condense_rank_profile",
    "remove_cand_rank_ballot",
    "condense_rank_ballot",
    "remove_repeat_cands_rank_ballot",
    "remove_cand_score_ballot",
    "clean_score_profile",
    "remove_cand_score_profile",
]



import os
import csv
import io
import urllib.request
from pathlib import Path
from pandas.errors import EmptyDataError, DataError
from typing import Union

from votekit.pref_profile import RankProfile
from votekit.ballot import RankBallot


def load_scottish(
    fpath: Union[str, os.PathLike, Path],
) -> tuple[RankProfile, int, list[str], dict[str, str], str]:
    """
    Given a file path, loads cast vote record from format used for Scottish election data
    in (this repo)[https://github.com/mggg/scot-elex].

    Args:
        fpath (Union[str, os.PathLike, pathlib.Path]): Path to Scottish election csv file.
            Can be a url.

    Raises:
        FileNotFoundError: If fpath is invalid.
        EmptyDataError: If dataset is empty.
        DataError: If there is missing or incorrect metadata or candidate data.

    Returns:
        tuple:
            A tuple ``(RankProfile, seats, cand_list, cand_to_party, ward)``
            representing the election, the number of seats in the election, the candidate
            names, a dictionary mapping candidates to their party, and the ward. The
            candidate names are also stored in the PreferenceProfile object.
    """

    # Convert the ballot rows to ints while leaving the candidates as strings
    def convert_row(row):
        return [int(item) if item.isdigit() else item for item in row]

    def parse_csv_reader(reader):
        data = []
        for row in reader:
            # This just removes any empty strings that are hanging out since
            # we don't need to preserve columns
            filtered_row = list(filter(lambda x: x != "", row))

            # only save non-empty rows
            if len(filtered_row) > 0:
                data.append(convert_row(filtered_row))
        return data

    fpath = str(fpath)

    if not os.path.isfile(fpath):
        with urllib.request.urlopen(fpath) as response:
            data = response.read().decode("utf-8")
        reader = csv.reader(io.StringIO(data))
        data = parse_csv_reader(reader)

    else:
        if os.path.getsize(fpath) == 0:
            raise EmptyDataError(f"CSV at {fpath} is empty.")
        with open(fpath, "r", encoding="utf-8") as f:
            reader = csv.reader(f)
            data = parse_csv_reader(reader)

    if len(data[0]) != 2:
        raise DataError(
            "The metadata in the first row should be number of \
                            candidates, seats."
        )

    cand_num, seats = data[0][0], data[0][1]
    ward = data[-1][0]

    num_to_cand = {}
    cand_to_party = {}

    data_cand_num = len([r for r in data if "Candidate" in str(r[0])])
    if data_cand_num != cand_num:
        raise DataError(
            "Incorrect number of candidates in either first row metadata \
                        or in candidate list at end of csv file."
        )

    # record candidate names, which are up until the final row
    for i, line in enumerate(data[len(data) - (cand_num + 1) : -1]):
        if "Candidate" not in line[0]:
            raise DataError(
                f"The number of candidates on line 1 is {cand_num}, which\
                            does not match the metadata."
            )
        cand = line[1]
        party = line[2]

        # candidates are 1 indexed
        num_to_cand[i + 1] = cand
        cand_to_party[cand] = party

    cand_list = list(cand_to_party.keys())

    ballots = [RankBallot()] * len(data[1 : len(data) - (cand_num + 1)])

    for i, line in enumerate(data[1 : len(data) - (cand_num + 1)]):
        ballot_weight = line[0]
        cand_ordering = line[1:]
        ranking = tuple([frozenset({num_to_cand[n]}) for n in cand_ordering])

        ballots[i] = RankBallot(ranking=ranking, weight=ballot_weight)

    profile = RankProfile(
        ballots=tuple(ballots), candidates=tuple(cand_list)
    ).group_ballots()
    return (profile, seats, cand_list, cand_to_party, ward)
